#!/usr/bin/env python
#
# Copyright:
# ----------------------------------------------------------------------------
# This confidential and proprietary software may be used only as authorized
# by a licensing agreement from ARM Limited.
#      (C) COPYRIGHT 2016-2018 ARM Limited, ALL RIGHTS RESERVED
# The entire notice above must be reproduced on all authorized copies and
# copies may only be made to the extent permitted by a licensing agreement
# from ARM Limited.
# ----------------------------------------------------------------------------

import argparse
import importlib
import logging
import os
import re
import sys

# This script is actually within our package, so add the package to the python path
sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
from config_system.general import enforce_dependent_values, get_config, init_config, \
    read_config_file, read_profile_file, set_config_if_prompt, write_config
import config_system.log_handlers as log_handlers

logging.basicConfig(format='%(levelname)s: %(message)s', level=logging.WARNING)
counter = log_handlers.ErrorCounterHandler()
logging.getLogger().addHandler(counter) # root logger

logger = logging.getLogger(__name__)

parser = argparse.ArgumentParser()
parser.add_argument('-o', '--output', required=True,
                    help='Path to the output file')
parser.add_argument('-d', '--database', default="Mconfig",
                    help='Path to the configuration database (Mconfig)')
parser.add_argument('-p', '--plugin', action='append',
                    help='Post configuration plugin to execute',
                    default=[])
parser.add_argument('--ignore-missing', action='store_true', default=False,
                    help="Ignore missing database files included with 'source'")
parser.add_argument('args', nargs="*")
args = parser.parse_args()

init_config(args.database, args.ignore_missing)

files = []
setters = []

CONFIG_ARG_RE = re.compile(r'^([A-Za-z_][A-Za-z0-9_]*)=(.*)$')
def parse_config_arg(arg):
    """ Parse a KEY=VALUE command-line argument """
    m = CONFIG_ARG_RE.match(arg)
    if m is None:
        return None, None
    else:
        return m.group(1), m.group(2)

for arg in args.args:
    key, value = parse_config_arg(arg)
    if key:
        set_config_if_prompt(key, value, True)
    else:
        logger.info("Reading %s" % arg)
        read_profile_file(arg)

enforce_dependent_values()

for arg in args.args:
    key, value = parse_config_arg(arg)
    if key:
        try:
            final_value = get_config(key)['value']
        except KeyError:
            logger.error("unknown configuration option \"%s\"" % key)
        else:
            if final_value != value:
                logger.error("%s=%s was ignored or overriden. Value is %s" %
                             (key, value, final_value))

for plugin in args.plugin:
    path, name = os.path.split(plugin)
    if path.startswith('/'):
        sys.path.insert(0, path)
    else:
        sys.path.insert(0, os.path.join(os.getcwd(), path))
    sys.path.append(os.path.dirname(sys.argv[0]))
    try:
        mod = importlib.import_module(name)
        mod.plugin_exec()
    except ImportError as err:
        logger.error(err)
    except Exception as err:
        logger.warning("Problem encountered in %s plugin: %s" % (name, repr(err)))
        import traceback
        traceback.print_tb(sys.exc_info()[2])

write_config(args.output)

issues = counter.errors() + counter.criticals()
sys.exit(0 if issues == 0 else 1)
