/*
 * Copyright:
 * ----------------------------------------------------------------------------
 * This confidential and proprietary software may be used only as authorized
 * by a licensing agreement from ARM Limited.
 *      (C) COPYRIGHT 2016-2018 ARM Limited, ALL RIGHTS RESERVED
 * The entire notice above must be reproduced on all authorized copies and
 * copies may only be made to the extent permitted by a licensing agreement
 * from ARM Limited.
 * ----------------------------------------------------------------------------
 */

package core

import (
	"fmt"
	"reflect"
	"strings"

	"github.com/google/blueprint/proptools"

	"bob/utils"
)

func titleFirst(a string) string {
	result := ""
	if len(a) > 0 {
		result += strings.ToUpper(a[0:1])
	}
	if len(a) > 1 {
		result += strings.ToLower(a[1:])
	}
	return result
}

// Generate and initialize a struct containing a field of type 'propsType' for
// every available feature. The generated object is embedded in each module
// types' properties instance, and is used by Blueprint to decide what
// properties can be set inside features in each module type.
//
// An example generated type:
// type featureSetType struct {
//         Debug PropsType
//         Enable_something PropsType
//         Some_other_feature PropsType
// }
func (f *Features) Init(config *Config, propsType reflect.Type) {
	fields := make([]reflect.StructField, len(config.Properties.Features))

	for i, key := range utils.SortedKeysBoolMap(config.Properties.Features) {
		field := reflect.StructField{
			Name: titleFirst(key),
			Type: propsType,
		}
		fields[i] = field
	}

	featureSetType := reflect.StructOf(fields)
	featureSetValPtr := reflect.New(featureSetType)
	f.BlueprintEmbed = featureSetValPtr.Interface()
}

// Merge properties from enabled features in src to dst
func AppendProps(dst []interface{}, src interface{}, properties *ConfigProperties) error {
	featureSetVal := reflect.ValueOf(src).Elem()

	for _, key := range utils.SortedKeysBoolMap(properties.Features) {
		// Check the feature is enabled
		if properties.Features[key] {
			field := featureSetVal.FieldByName(titleFirst(key))

			if !field.IsValid() {
				panic(fmt.Sprintf("Field returned for property %s isn't valid\n", key))
			}

			// AppendProperties expects a pointer to a struct.
			featureProps := field.Addr().Interface()

			// If featureProps is nil then we've determined that we can skip this,
			// so avoid calling AppendProperties
			if featureProps != nil {
				err := proptools.AppendMatchingProperties(dst, featureProps, nil)
				if err != nil {
					return err
				}
			}
		}
	}
	return nil
}
