/*
 * Copyright:
 * ----------------------------------------------------------------------------
 * This confidential and proprietary software may be used only as authorized
 * by a licensing agreement from ARM Limited.
 *      (C) COPYRIGHT 2017-2018 ARM Limited, ALL RIGHTS RESERVED
 * The entire notice above must be reproduced on all authorized copies and
 * copies may only be made to the extent permitted by a licensing agreement
 * from ARM Limited.
 * ----------------------------------------------------------------------------
 */

package core

import (
	"errors"
	"path/filepath"
	"reflect"
	"regexp"
	"strings"

	"github.com/google/blueprint"

	"bob/utils"
)

var (
	generatedHeaderTag = DependencyTag{name: "generated_headers"}
	generatedSourceTag = DependencyTag{name: "generated_sources"}
	generatedDepTag    = DependencyTag{name: "generated_dep"}
	hostToolBinTag     = DependencyTag{name: "host_tool_bin"}
)

// For bob_transform_source each src in the glob will get its own
// ninja rule. Each src can have multiple outputs.
//
// To support this the Inout type is used to group the outputs
// associated with each src.
//
// Inputs are relative to the project directory (they include module
// directory but not base source directory). Generated inputs are
// therefore kept separately in gen_in.
type Inout struct {
	src_in        []string
	gen_in        []string
	out           []string
	depfile       string
	implicit_srcs []string
}

type GenerateProps struct {
	SourceProps
	AliasableProps
	EnableableProps
	InstallableProps

	/* The command that is to be run for this source generation.
	 * Substitutions can be made in the command, by using $name_of_var. A list of substitutions that can be used:
	 * $gen_dir    - the path to the directory which belongs to this source generator
	 * $in         - the path to the sources - space-delimited
	 * $out        - the path to the targets - space-delimited
	 * $args       - the value of "args" - space-delimited
	 * $tool       - the path to the tool
	 * $host_bin   - the path to the binary that is produced by the host_bin module
	 * $(name)_dir - the build directory for each dep in generated_dep
	 * $src_dir    - the path to the driver source directory - this will be different than the build source directory
	 *               for Android.
	 * $module_dir - the path to the module directory */
	Cmd string

	// A path to the tool that is to be used in cmd. If $tool is in the command variable, then this will be replaced.
	// with the path to this tool
	Tool string

	// A path to a binary that is to be
	Host_bin string

	// Values to use on Android for LOCAL_MODULE_TAGS, defining which builds this module is built for
	// TODO: Hide this in Android-specific properties
	Tags []string

	// A list of other modules that this generator depends on. The dependencies can be used in the command through
	// $name_of_dependency_dir .
	Module_deps []string

	// A list of other modules that this generator depends on. The dependencies will be add to the list of srcs
	Module_srcs []string

	// A list of args that will be spaceseparated and add to the cmd
	Args []string

	// Used to indicate that the console should be used.
	Console bool

	// Additional include paths to add for modules that use generate_headers.
	// This will be defined relative to the module-specific build directory
	Export_gen_include_dirs []string

	// The defaults used to retrieve cflags
	Flag_defaults []string

	// The target type - must be either "host" or "target"
	Target string
}

type GenerateCommon struct {
	blueprint.SimpleName
	Properties struct {
		GenerateProps
		Features
		FlagArgsBuild Build `blueprint:"mutated"`
	}
}

type HostBin interface {
	HostBin() string
}

type GeneratedDepender interface {
	Build() *Build
}

// When referencing libraries provided by a generator module use "module/path/to/lib"
// This function splits the reference into the module and the library.
func splitGeneratedComponent(comp string) (module string, lib string) {
	split := strings.Split(comp, "/")

	if len(split) < 2 {
		panic(errors.New("Generated component " + comp + " does not specify module and lib"))
	}

	return split[0], strings.Join(split[1:], "/")
}

func (m *GenerateCommon) OutputDir(g GeneratorBackend) string {
	return g.sourceOutputDir(m)
}

func (m *GenerateCommon) ShortName() string {
	return m.Name()
}

func (m *GenerateCommon) AltName() string {
	return m.Name()
}

func (m *GenerateCommon) AltShortName() string {
	return m.ShortName()
}

// Workaround for Golang not having a way of querying superclasses

func (m *GenerateCommon) GetGenerateCommon() *GenerateCommon {
	return m
}

func (m *GenerateCommon) topLevelProperties() []interface{} {
	return []interface{}{&m.Properties.GenerateProps}
}

func (m *GenerateCommon) Features() *Features {
	return &m.Properties.Features
}

func (m *GenerateCommon) getTarget() string {
	return m.Properties.Target
}

func (m *GenerateCommon) GetInstallableProps() *InstallableProps {
	return &m.Properties.InstallableProps
}

func (m *GenerateCommon) GetInstallDepPhonyNames(ctx blueprint.ModuleContext) []string {
	return GetShortNamesForDirectDepsWithTags(ctx, InstallDepTag)
}

// Return a list of headers generated by this module with full paths
func getHeadersGenerated(g GeneratorBackend, m DependentInterface) []string {
	return utils.Filter(m.Outputs(g), utils.IsHeader)
}

// Return a list of source files (not headers) generated by this module with full paths
func getSourcesGenerated(g GeneratorBackend, m DependentInterface) []string {
	return utils.Filter(m.Outputs(g), utils.IsSource)
}

type GetGenerateCommonInterface interface {
	GetGenerateCommon() *GenerateCommon
}

func GetGenerateCommon(i interface{}) (*GenerateCommon, bool) {
	var gsc *GenerateCommon
	gsd, ok := i.(GetGenerateCommonInterface)
	if ok {
		gsc = gsd.GetGenerateCommon()
	}
	return gsc, ok
}

func (m *GenerateCommon) GetEnableableProps() *EnableableProps {
	return &m.Properties.EnableableProps
}

// The module that can generate sources using a single execution
// The command will be run once - with $in being the paths in "srcs" and $out being the paths in "out".
// The working directory will be the source directory, and all paths will be relative to the source directory
// if not else noted.

type GenerateSourceProps struct {
	// The list of files that will be output.
	Out []string
	// File generated by the command describing implicit dependencies.
	Depfile string
	// List of implicit sources not described by the depfile
	Implicit_srcs []string
}

type GenerateSource struct {
	GenerateCommon
	Properties struct {
		GenerateSourceProps
	}
}

func (m *GenerateSource) Outputs(g GeneratorBackend) []string {
	return utils.PrefixDirs(m.Properties.Out, m.OutputDir(g))
}

func (m *GenerateSource) GenerateBuildActions(ctx blueprint.ModuleContext) {
	if IsEnabled(m) {
		inouts := m.Inouts(ctx)
		getBackend(ctx).GenerateSourceActions(m, ctx, inouts)
	}
}

func (m *GenerateSource) topLevelProperties() []interface{} {
	return []interface{}{&m.Properties.GenerateSourceProps, &m.GenerateCommon.Properties.GenerateProps}
}

// Returns the tool binary for a GenerateSource module. This is different from the "tool"
// in that it used to depend on a bob_binary module
func (m *GenerateCommon) getHostBin(ctx blueprint.ModuleContext) (string, string) {
	g := getBackend(ctx)
	toolBin := ""
	toolTarget := ""

	ctx.VisitDirectDepsIf(
		func(m blueprint.Module) bool {
			return ctx.OtherModuleDependencyTag(m) == hostToolBinTag
		},
		func(m blueprint.Module) {
			if b, ok := m.(*Binary); ok {
				toolBin = b.Outputs(g)[0]
				toolTarget = b.getTarget()
			} else if b, ok := m.(*GenerateBinary); ok {
				toolBin = b.Outputs(g)[0]
			} else {
				panic(errors.New("bob_generate_source must depend on a binary, not a library"))
			}
		})
	return toolBin, toolTarget
}

// Returns the dependents of a GenerateSource module. This is used for more complex dependencies, where
// the dependencies are not just a binary or a headers, but where the paths are used directly in a script
func getDependentArgsAndFiles(ctx blueprint.ModuleContext, args map[string]string) (depfiles []string) {
	g := getBackend(ctx)
	ctx.VisitDirectDepsIf(
		func(m blueprint.Module) bool {
			return ctx.OtherModuleDependencyTag(m) == generatedDepTag
		},
		func(m blueprint.Module) {
			gen, ok := m.(DependentInterface)
			if !ok {
				panic(errors.New(reflect.TypeOf(m).String() + " is not a valid dependent interface"))
			}

			depName := ctx.OtherModuleName(m)
			args[depName+"_dir"] = gen.OutputDir(g)
			args[depName+"_out"] = strings.Join(gen.Outputs(g), " ")
			depfiles = append(depfiles, gen.Outputs(g)...)
		})

	return
}

func (m *GenerateCommon) GetArgs(ctx blueprint.ModuleContext) (string, map[string]string, []string, string) {
	g := getBackend(ctx)

	cc, cctargetflags := GetCCompiler(ctx, m.Properties.Target)
	cxx, cxxtargetflags := GetCXXCompiler(ctx, m.Properties.Target)
	props := &m.Properties.FlagArgsBuild

	args := map[string]string{
		"ar":                GetArchiver(ctx, m.Properties.Target),
		"bob_config":        filepath.Join(g.buildDir(), configName),
		"bob_config_opts":   configOpts,
		"cc":                cc,
		"cflags":            strings.Join(props.Cflags, " "),
		"conlyflags":        cctargetflags + " " + strings.Join(props.Conlyflags, " "),
		"cxx":               cxx,
		"cxxflags":          cxxtargetflags + " " + strings.Join(props.Cxxflags, " "),
		"gen_dir":           g.sourceOutputDir(m),
		"headers_generated": "",
		"module_dir":        filepath.Join(g.sourcePrefix(), ctx.ModuleDir()),
		"shared_libs_dir":   g.sharedLibsDir(m.Properties.GenerateProps.Target),
		"src_dir":           g.sourcePrefix(),
		"srcs_generated":    "",
		"sysroot":           getConfig(ctx).Properties.GetString("clang_sysroot"),
	}

	args["build_wrapper"], _ = props.getBuildWrapperAndDeps(ctx)

	dependents := getDependentArgsAndFiles(ctx, args)

	if m.Properties.Tool != "" {
		tool_path := filepath.Join(g.sourcePrefix(), ctx.ModuleDir(), m.Properties.Tool)
		args["tool"] = tool_path
		dependents = append(dependents, tool_path)
	}

	host_bin, host_target := m.getHostBin(ctx)
	if host_bin != "" {
		args["host_bin"] = host_bin
		dependents = append(dependents, host_bin)
	}

	// Args can contain other parameters, so replace that immediately
	cmd := strings.Replace(m.Properties.Cmd, "${args}", strings.Join(m.Properties.Args, " "), -1)

	return cmd, args, dependents, host_target
}

func (m *GenerateCommon) getSources(ctx blueprint.ModuleContext) []string {
	return m.Properties.GetSrcs(ctx)
}

func (m *GenerateCommon) processPaths(ctx blueprint.BaseModuleContext) {
	m.Properties.SourceProps.processPaths(ctx)
}

func (m *GenerateCommon) getAliasList() []string {
	return m.Properties.getAliasList()
}

func (m *GenerateSource) Inouts(ctx blueprint.ModuleContext) []Inout {
	var inout Inout
	g := getBackend(ctx)
	inout.src_in = m.GenerateCommon.Properties.GetSrcs(ctx)
	inout.gen_in = append(m.GenerateCommon.Properties.SourceProps.Specials, getGeneratedFiles(ctx)...)
	inout.out = m.Outputs(g)
	if m.Properties.Depfile != "" {
		inout.depfile = filepath.Join(g.sourceOutputDir(&m.GenerateCommon), m.Properties.Depfile)
	}
	inout.implicit_srcs = utils.PrefixDirs(m.Properties.Implicit_srcs, filepath.Join(g.sourcePrefix(), ctx.ModuleDir()))

	return []Inout{inout}
}

func (m *GenerateSource) FilesToInstall(ctx blueprint.ModuleContext) []string {
	// Install everything that we generate
	g := getBackend(ctx)
	return m.Outputs(g)
}

type TransformSourceProperties struct {
	GenerateCommon
	Properties struct {
		// The regular expression that is used to transform the source path to the target path.
		Out struct {
			// Regular expression to capture groups from srcs
			Match string
			// Names of outputs, which can use capture groups from match
			Replace []string
			// Name of the dependency file (if needed), which can use capture groups from match
			Depfile string
			// List of implicit sources not described by the depfile
			Implicit_srcs []string
		}
	}
}

func (m *TransformSource) Outputs(g GeneratorBackend) []string {
	return m.outputs
}

func (m *TransformSource) GenerateBuildActions(ctx blueprint.ModuleContext) {
	if IsEnabled(m) {
		inouts := m.Inouts(ctx)
		for _, inout := range inouts {
			m.outputs = append(m.outputs, inout.out...)
		}
		getBackend(ctx).TransformSourceActions(m, ctx, inouts)
	}
}

func (m *TransformSource) Inouts(ctx blueprint.ModuleContext) []Inout {
	var inouts []Inout
	g := getBackend(ctx)
	re := regexp.MustCompile(m.Properties.Out.Match)

	// For a transform source every input file is expected to be
	// converted to an output file. So each file in srcs will get its
	// own inout and each generated input (from module_srcs) will also
	// get its own inout. That means one of src_in or gen_in is always
	// empty, and the other has a single element.
	empty := []string{}

	for _, src := range m.GenerateCommon.Properties.GetSrcs(ctx) {
		ins := []string{src}
		outs := []string{}
		depfile := ""
		implicit_srcs := []string{}

		for _, rep := range m.Properties.Out.Replace {
			out := filepath.Join(g.sourceOutputDir(&m.GenerateCommon), re.ReplaceAllString(src, rep))
			outs = append(outs, out)
		}
		if m.Properties.Out.Depfile != "" {
			depfile = filepath.Join(g.sourceOutputDir(&m.GenerateCommon),
				re.ReplaceAllString(src, m.Properties.Out.Depfile))
		}

		for _, impl_src := range m.Properties.Out.Implicit_srcs {
			impl_src = re.ReplaceAllString(src, impl_src)
			impl_src = filepath.Join(g.sourcePrefix(), ctx.ModuleDir(), impl_src)
			implicit_srcs = append(implicit_srcs, impl_src)
		}

		inouts = append(inouts, Inout{ins, empty, outs, depfile, implicit_srcs})
	}
	for _, src := range append(m.GenerateCommon.Properties.Specials, getGeneratedFiles(ctx)...) {
		ins := []string{src}
		outs := []string{}
		depfile := ""
		implicit_srcs := []string{}

		for _, rep := range m.Properties.Out.Replace {
			out := filepath.Join(g.sourceOutputDir(&m.GenerateCommon), re.ReplaceAllString(src, rep))
			outs = append(outs, out)
		}

		if m.Properties.Out.Depfile != "" {
			depfile = filepath.Join(g.sourceOutputDir(&m.GenerateCommon),
				re.ReplaceAllString(src, m.Properties.Out.Depfile))
		}

		for _, impl_src := range m.Properties.Out.Implicit_srcs {
			impl_src = re.ReplaceAllString(src, impl_src)
			impl_src = filepath.Join(g.sourcePrefix(), ctx.ModuleDir(), impl_src)
			implicit_srcs = append(implicit_srcs, impl_src)
		}

		inouts = append(inouts, Inout{empty, ins, outs, depfile, implicit_srcs})
	}

	return inouts
}

func (m *TransformSource) FilesToInstall(ctx blueprint.ModuleContext) []string {
	// Install everything that we generate
	g := getBackend(ctx)
	return m.Outputs(g)
}

// The module that can generate sources using a multiple execution
// The command will be run once per src file- with $in being the path in "srcs" and $out being the path transformed
// through the regexp defined by out.match and out.replace. The regular expression that is used is
// in regexp.compiled(out.Match).ReplaceAllString(src[i], out.Replace). See https://golang.org/pkg/regexp/ for more
// information.
// The working directory will be the source directory, and all paths will be relative to the source directory
// if not else noted
type TransformSource struct {
	TransformSourceProperties
	outputs []string
}

func GenerateSourceFactory(config *Config) (blueprint.Module, []interface{}) {
	module := &GenerateSource{}
	module.GenerateCommon.Properties.Features.Init(config, reflect.TypeOf(GenerateProps{}))
	return module, []interface{}{&module.GenerateCommon.Properties, &module.Properties,
		&module.SimpleName.Properties}
}

func TransformSourceFactory(config *Config) (blueprint.Module, []interface{}) {
	module := &TransformSource{}
	module.GenerateCommon.Properties.Features.Init(config, reflect.TypeOf(GenerateProps{}))
	return module, []interface{}{&module.GenerateCommon.Properties,
		&module.Properties,
		&module.SimpleName.Properties}
}

// ModuleContext Helpers

// Return the Outputs() of all GeneratedSource dependencies of the current
// module. The current module can be generated or a library, and the
// dependencies can be anything implementing DependentInterface (so "generated"
// is a misnomer, because this includes libraries, too).
func getGeneratedFiles(ctx blueprint.ModuleContext) []string {
	var srcs []string
	g := getBackend(ctx)
	ctx.VisitDirectDepsIf(
		func(m blueprint.Module) bool { return ctx.OtherModuleDependencyTag(m) == generatedSourceTag },
		func(m blueprint.Module) {
			if gs, ok := m.(DependentInterface); ok {
				srcs = append(srcs, gs.Outputs(g)...)
			} else {
				panic(errors.New(ctx.OtherModuleName(m) + " does not have outputs"))
			}
		})
	return srcs
}

func generatedDependerMutator(mctx blueprint.BottomUpMutatorContext) {
	if e, ok := mctx.Module().(Enableable); ok {
		if !IsEnabled(e) {
			// Not enabled, so don't add dependencies
			return
		}
	}

	// Things which depend on generated/transformed sources
	if gd, ok := mctx.Module().(GeneratedDepender); ok {
		if _, ok := mctx.Module().(*Defaults); ok {
			// We do not want to add dependencies for defaults
			return
		}
		b := gd.Build()
		mctx.AddDependency(mctx.Module(), generatedSourceTag, b.Generated_sources...)
		mctx.AddDependency(mctx.Module(), generatedHeaderTag, b.Generated_headers...)
		mctx.AddDependency(mctx.Module(), generatedDepTag, b.Generated_deps...)
	}

	// Things that a generated/transformed source depends on
	if gsc, ok := GetGenerateCommon(mctx.Module()); ok {
		if gsc.Properties.Host_bin != "" {
			parseAndAddVariationDeps(mctx, hostToolBinTag,
				gsc.Properties.Host_bin)
		}
		// Generated sources can use the outputs of another generated
		// source or library as a source file or dependency.
		parseAndAddVariationDeps(mctx, generatedDepTag,
			gsc.Properties.Module_deps...)
		parseAndAddVariationDeps(mctx, generatedSourceTag,
			gsc.Properties.Module_srcs...)
	}
}
