/*
 * Copyright:
 * ----------------------------------------------------------------------------
 * This confidential and proprietary software may be used only as authorized
 * by a licensing agreement from ARM Limited.
 *      (C) COPYRIGHT 2017-2018 ARM Limited, ALL RIGHTS RESERVED
 * The entire notice above must be reproduced on all authorized copies and
 * copies may only be made to the extent permitted by a licensing agreement
 * from ARM Limited.
 * ----------------------------------------------------------------------------
 */

package core

import (
	"errors"
	"fmt"
	"path/filepath"
	"reflect"
	"strings"

	"github.com/google/blueprint"

	"bob/graph"
	"bob/utils"
)

type BuildProps struct {
	SourceProps
	AliasableProps

	// Alternate output name, used for the file name and Android rules
	Out string

	// Flags used for C/C++ compilation
	Cflags []string
	// Flags exported for dependent modules
	Export_cflags []string
	// Flags used for C compilation
	Conlyflags []string
	// Flags used for C++ compilation
	Cxxflags []string
	// Flags used for assembly compilation
	Asflags []string
	// Flags used for linking
	Ldflags []string
	// Same as ldflags, but specified on static libraries and propagated to
	// the top-level build object.
	Export_ldflags []string
	// Shared library version
	Library_version string

	// The list of shared lib modules that this library depends on.
	// 'shared_libs' only makes sense for share libraries and executables -
	// static libs should use 'export_shared_libs', so the dependency is
	// propagated to the top-level object.
	Shared_libs []string
	// The libraries mentioned here will be appended to shared_libs of the modules that use
	// this library (via static_libs, whole_static_libs or shared_libs).
	// export_shared_libs is an indication that this module is using a shared library, and
	// users of this module need to link against it.
	Export_shared_libs []string
	Extra_shared_libs  []string `blueprint:"mutated"`

	// The list of static lib modules that this library depends on
	// Only makes sense to define for shared libraries and executables
	Static_libs []string
	// The libraries mentioned here will be appended to static_libs of the modules that use
	// this library (via static_libs, whole_static_libs or shared_libs).
	// export_static_libs is an indication that this module is using a static library, and
	// users of this module need to link it.
	Export_static_libs []string

	// This list of dependencies that exported cflags and exported include dirs
	// should be propagated 1-level higher
	Reexport_libs []string
	// Internal property for collecting libraries with reexported flags and include paths
	Resolved_reexported_libs []string `blueprint:"mutated"`

	Resolved_static_libs []string `blueprint:"mutated"`

	// The list of whole static libraries that this library depnds on
	// This will include all the objects in the library (as opposed to normal static linking)
	// If this is set for a static library, any shared library will also include objects
	// from dependent libraries
	Whole_static_libs []string

	// Linker flags required to link to the necessary system libraries
	Ldlibs []string
	// Same as ldlibs, but specified on static libraries and
	// propagated to the top-level build object.
	Export_ldlibs []string

	// The list of modules that generate extra headers for this module
	Generated_headers []string

	// The list of modules that generate extra source files for this module
	Generated_sources []string

	// The list of modules that generate output required by the build wrapper
	Generated_deps []string

	// Values to use on Android for LOCAL_MODULE_TAGS, defining which builds this module is built for
	// TODO: Hide this in Android-specific properties
	Tags []string

	// Value to use on Android for LOCAL_MODULE_OWNER
	// TODO: Hide this in Android-specific properties
	Owner string

	// The list of include dirs to use that is relative to the source directory
	Include_dirs []string

	// The list of include dirs to use that is relative to the build.bp file
	Local_include_dirs []string // These use relative instead of absolute paths

	// Include local dirs to be exported into dependent
	Export_local_include_dirs []string

	// Include dirs (path relative to DDK root) to be exported into dependent
	Export_include_dirs []string // TODO: Hide this in Android-specific properties

	// Wrapper for all build commands (object file compilation *and* linking)
	Build_wrapper *string
	// Files in the source directory that the wrapper depends on.
	Build_wrapper_deps []string

	// This is a shared library that pulls in another shared library which will resolve symbols that
	// the binary needs. Only valid on bob_shared_library.
	//
	// Currently we need to link this with -Wl,--copy-dt-needed-entries,
	// but this makes the binary depend on the forwarded library too
	Forwarding_shlib *bool

	InstallableProps
	EnableableProps
	SplittableProps

	// Linux kernel config options to emulate. These are passed to Kbuild in
	// the 'make' command-line, and set in the source code via EXTRA_CFLAGS
	Kbuild_options []string
	// Kernel modules which this module depends on
	Extra_symbols []string
	// Arguments to pass to kernel make invocation
	Make_args []string
	// Kernel directory location
	Kernel_dir string
	// Compiler prefix for kernel build
	Kernel_compiler string

	Target_type string `blueprint:"mutated"`
}

type Build struct {
	BuildProps
	Target TargetSpecific
	Host   TargetSpecific
}

// These function check the boolean pointers - which are only filled if someone sets them
// If not, the default value is returned

func (l *Build) isHostSupported() bool {
	if l.Host_supported == nil {
		return false
	}
	return *l.Host_supported
}

func (l *Build) isTargetSupported() bool {
	if l.Target_supported == nil {
		return true
	}
	return *l.Target_supported
}

func (l *Build) IsForwardingSharedLibrary() bool {
	if l.Forwarding_shlib == nil {
		return false
	}
	return *l.Forwarding_shlib
}

func (l *Build) getBuildWrapperAndDeps(ctx blueprint.ModuleContext) (string, []string) {
	if l.Build_wrapper != nil {
		depargs := map[string]string{}
		files := getDependentArgsAndFiles(ctx, depargs)

		// Replace any property usage in buildWrapper
		buildWrapper := *l.Build_wrapper
		for k, v := range depargs {
			buildWrapper = strings.Replace(buildWrapper, "${"+k+"}", v, -1)
		}

		return buildWrapper, append(l.Build_wrapper_deps, files...)
	}

	return "", []string{}
}

// Add module paths to srcs, exclude_srcs, local_include_dirs and export_local_include_dirs
func (l *Build) processPaths(ctx blueprint.BaseModuleContext) {
	prefix := ctx.ModuleDir()
	l.SourceProps.processPaths(ctx)
	l.Local_include_dirs = utils.PrefixDirs(l.Local_include_dirs, prefix)
	l.Export_local_include_dirs = utils.PrefixDirs(l.Export_local_include_dirs, prefix)

	// When prefixPaths is called we have collapsed features, but not
	// targets, so we also need to expand paths in host and target
	// specific properties as well.
	l.Host.SourceProps.processPaths(ctx)
	l.Host.Local_include_dirs = utils.PrefixDirs(l.Host.Local_include_dirs, prefix)
	l.Host.Export_local_include_dirs = utils.PrefixDirs(l.Host.Export_local_include_dirs, prefix)

	l.Target.SourceProps.processPaths(ctx)
	l.Target.Local_include_dirs = utils.PrefixDirs(l.Target.Local_include_dirs, prefix)
	l.Target.Export_local_include_dirs = utils.PrefixDirs(l.Target.Export_local_include_dirs, prefix)
}

// A base class for different library modules
type Library struct {
	blueprint.SimpleName
	Properties struct {
		Features
		Build
		// The list of default properties that should prepended to all configuration
		Defaults []string
	}
}

func (l *Library) Defaults() []string {
	return l.Properties.Defaults
}

func (l *Library) Build() *Build {
	return &l.Properties.Build
}

func (l *Library) topLevelProperties() []interface{} {
	return []interface{}{&l.Properties.Build.BuildProps}
}

func (l *Library) Features() *Features {
	return &l.Properties.Features
}

func (l *Library) getTarget() string {
	return l.Properties.Target_type
}

func (l *Library) GetInstallableProps() *InstallableProps {
	return &l.Properties.InstallableProps
}

func (l *Library) GetInstallDepPhonyNames(ctx blueprint.ModuleContext) []string {
	return GetShortNamesForDirectDepsWithTags(ctx, InstallDepTag, SharedDepTag)
}

func (l *Library) GetEnableableProps() *EnableableProps {
	return &l.Properties.Build.EnableableProps
}

func (l *Library) getAliasList() []string {
	return l.Properties.getAliasList()
}

func (l *Library) supportedVariants() (targets []string) {
	if l.Properties.isHostSupported() {
		targets = append(targets, HOST_TAG)
	}
	if l.Properties.isTargetSupported() {
		targets = append(targets, TARGET_TAG)
	}
	return
}

func (l *Library) disable() {
	f := false
	l.Properties.Enabled = &f
}

func (l *Library) setVariant(tgt string) {
	l.Properties.Target_type = tgt
}

func (l *Library) getSplittableProps() *SplittableProps {
	return &l.Properties.SplittableProps
}

func (l *Library) OutputName() string {
	if len(l.Properties.Out) > 0 {
		return l.Properties.Out
	}
	return l.SimpleName.Name()
}

func (l *Library) AltName() string {
	return l.OutputName()
}

func (l *Library) AltShortName() string {
	if len(l.supportedVariants()) > 1 {
		return l.AltName() + "__" + l.Properties.Target_type
	}
	return l.AltName()
}

// Returns the shortname for the output, which is used as a phony target. If it
// can be built for multiple variants, require a '__host' or '__target' suffix to
// disambiguate.
func (l *Library) ShortName() string {
	if len(l.supportedVariants()) > 1 {
		return l.Name() + "__" + l.Properties.Target_type
	}
	return l.Name()
}

func (l *Library) GetGeneratedHeaders(ctx blueprint.ModuleContext) (include_dirs []string, order_only []string) {
	g := getBackend(ctx)
	visited := map[string]bool{}
	ctx.VisitDirectDepsIf(
		func(m blueprint.Module) bool {
			tag := ctx.OtherModuleDependencyTag(m)
			return tag == generatedHeaderTag || tag == StaticDepTag || tag == SharedDepTag
		},
		func(m blueprint.Module) {

			if gs, ok := GetGenerateCommon(m); ok {
				// VisitDirectDepsIf will visit a module once for each
				// dependency. Only list the headers once.
				if _, ok := visited[m.Name()]; ok {
					return
				}

				include_dirs = append(include_dirs,
					utils.PrefixDirs(gs.Properties.Export_gen_include_dirs,
						g.sourceOutputDir(gs))...)
				// Generated headers are "order-only". That means that a source file does not need to rebuild
				// if a generated header changes, just that it must be built after a generated header.
				// The source file _will_ be rebuilt if it uses the header (since that is registered in the
				// depfile). Note that this means that generated headers cannot change which headers are used
				// (by aliasing another header).
				ds, ok := m.(DependentInterface)
				if !ok {
					panic(errors.New("Generated header must have Outputs()"))
				}
				generated_headers := getHeadersGenerated(g, ds)

				order_only = append(order_only, generated_headers...)
			}
		})
	return
}

func (l *Library) GetExportedVariables(ctx blueprint.ModuleContext) (exported_includes, exported_cflags []string) {
	visited := map[string]bool{}
	ctx.VisitDirectDeps(func(dep blueprint.Module) {

		if ctx.OtherModuleDependencyTag(dep) == WholeStaticDepTag ||
			ctx.OtherModuleDependencyTag(dep) == StaticDepTag ||
			ctx.OtherModuleDependencyTag(dep) == SharedDepTag ||
			ctx.OtherModuleDependencyTag(dep) == FlagDepTag {

			if _, ok := visited[dep.Name()]; ok {
				// VisitDirectDeps will visit a module once for each
				// dependency. We've already done this module.
				return
			}

			switch lib := dep.(type) {
			case *StaticLibrary:
				// Static libraries dependencies are in reversed order
				// so prepend to match the specified order
				exported_includes = append(exported_includes, lib.Properties.Export_include_dirs...)
				exported_includes = append(exported_includes, lib.Properties.Export_local_include_dirs...)
				exported_cflags = append(exported_cflags, lib.Properties.Export_cflags...)

			case *SharedLibrary:
				exported_includes = append(exported_includes,
					lib.Properties.Export_local_include_dirs...)
				exported_includes = append(exported_includes,
					lib.Properties.Export_include_dirs...)
				exported_cflags = append(exported_cflags,
					lib.Properties.Export_cflags...)
			}
		}
	})

	return
}

func (l *Library) processPaths(ctx blueprint.BaseModuleContext) {
	l.Properties.Build.processPaths(ctx)
}

func GetLibrary(i interface{}) (*Library, bool) {
	v := reflect.Indirect(reflect.ValueOf(i))
	field := v.FieldByName("Library")
	ok := false
	var l *Library
	if field.IsValid() {
		l, ok = field.Addr().Interface().(*Library)
	}
	return l, ok
}

type StaticLibrary struct {
	Library
}

func (m *StaticLibrary) OutputDir(g GeneratorBackend) string {
	return g.staticLibOutputDir(m)
}

func (m *StaticLibrary) Outputs(g GeneratorBackend) []string {
	return []string{filepath.Join(m.OutputDir(g), m.OutputName()+".a")}
}

func (m *StaticLibrary) FilesToInstall(ctx blueprint.ModuleContext) []string {
	return m.Outputs(getBackend(ctx))
}

func (l *Library) checkField(cond bool, fieldName string) {
	if !cond {
		panic(fmt.Sprintf("%s has field %s set", l.Name(), fieldName))
	}
}

func (m *StaticLibrary) GenerateBuildActions(ctx blueprint.ModuleContext) {
	props := m.Properties
	m.checkField(len(props.Shared_libs) == 0, "shared_libs")
	m.checkField(len(props.Static_libs) == 0, "static_libs")
	m.checkField(props.Forwarding_shlib == nil, "forwarding_shlib")

	if IsEnabled(m) {
		getBackend(ctx).StaticActions(m, ctx)
	}
}

type SharedLibrary struct {
	Library
}

func (m *SharedLibrary) getLinkName() string {
	return m.OutputName() + ".so"
}

func (m *SharedLibrary) getSoname() string {
	name := m.getLinkName()
	if m.Library.Properties.Library_version != "" {
		var v = strings.Split(m.Library.Properties.Library_version, ".")
		name += "." + v[0]
	}
	return name
}

func (m *SharedLibrary) getRealName() string {
	name := m.getLinkName()
	if m.Library.Properties.Library_version != "" {
		name += "." + m.Library.Properties.Library_version
	}
	return name
}

func (m *SharedLibrary) OutputDir(g GeneratorBackend) string {
	return g.sharedLibOutputDir(m)
}

func (m *SharedLibrary) Outputs(g GeneratorBackend) []string {
	if m.Library.Properties.Library_version == "" {
		return []string{filepath.Join(m.OutputDir(g), m.OutputName()+".so")}
	}
	return []string{filepath.Join(m.OutputDir(g), m.getRealName())}
}

func (m *SharedLibrary) FilesToInstall(ctx blueprint.ModuleContext) []string {
	return m.Outputs(getBackend(ctx))
}

func (m *SharedLibrary) LibrarySymlinks(ctx blueprint.ModuleContext) map[string]string {
	symlinks := map[string]string{}

	if m.Library.Properties.Library_version != "" {
		// To build you need a symlink from the link name and soname.
		// At runtime only the soname symlink is required.
		soname := m.getSoname()
		realName := m.getRealName()
		if soname == realName {
			panic(fmt.Errorf("module %s has invalid library_version '%s'",
				m.Name(),
				m.Library.Properties.Library_version))
		}
		symlinks[m.getLinkName()] = soname
		symlinks[soname] = realName
	}

	return symlinks
}

func (m *SharedLibrary) GenerateBuildActions(ctx blueprint.ModuleContext) {
	props := m.Properties
	m.checkField(len(props.Export_ldflags) == 0, "export_ldflags")
	m.checkField(len(props.Export_shared_libs) == 0, "export_shared_libs")
	m.checkField(len(props.Export_static_libs) == 0, "export_static_libs")
	m.checkField(len(props.Export_ldlibs) == 0, "export_ldlibs")

	if IsEnabled(m) {
		getBackend(ctx).SharedActions(m, ctx)
	}
}

type Binary struct {
	Library
}

func (m *Binary) OutputDir(g GeneratorBackend) string {
	return g.binaryOutputDir(m)
}

func (m *Binary) Outputs(g GeneratorBackend) []string {
	return []string{filepath.Join(m.OutputDir(g), m.OutputName())}
}

func (m *Binary) FilesToInstall(ctx blueprint.ModuleContext) []string {
	return m.Outputs(getBackend(ctx))
}

func (m *Binary) GenerateBuildActions(ctx blueprint.ModuleContext) {
	props := m.Properties
	m.checkField(len(props.Export_cflags) == 0, "export_cflags")
	m.checkField(len(props.Export_include_dirs) == 0, "export_include_dirs")
	m.checkField(len(props.Export_ldflags) == 0, "export_ldflags")
	m.checkField(len(props.Export_local_include_dirs) == 0, "export_local_include_dirs")
	m.checkField(len(props.Export_shared_libs) == 0, "export_shared_libs")
	m.checkField(len(props.Export_static_libs) == 0, "export_static_libs")
	m.checkField(len(props.Reexport_libs) == 0, "reexport_libs")
	m.checkField(len(props.Export_ldlibs) == 0, "export_ldlibs")
	m.checkField(len(props.Whole_static_libs) == 0, "whole_static_libs")
	m.checkField(props.Forwarding_shlib == nil, "forwarding_shlib")

	if IsEnabled(m) {
		getBackend(ctx).BinaryActions(m, ctx)
	}
}

func (l *Library) LibraryFactory(config *Config, module blueprint.Module) (blueprint.Module, []interface{}) {
	l.Properties.Features.Init(config, reflect.TypeOf(BuildProps{}))
	l.Properties.Build.Host.Features.Init(config, reflect.TypeOf(BuildProps{}))
	l.Properties.Build.Target.Features.Init(config, reflect.TypeOf(BuildProps{}))

	return module, []interface{}{&l.Properties, &l.SimpleName.Properties}
}

func StaticLibraryFactory(config *Config) (blueprint.Module, []interface{}) {
	module := &StaticLibrary{}
	return module.LibraryFactory(config, module)
}

func SharedLibraryFactory(config *Config) (blueprint.Module, []interface{}) {
	module := &SharedLibrary{}
	return module.LibraryFactory(config, module)
}

func BinaryFactory(config *Config) (blueprint.Module, []interface{}) {
	module := &Binary{}
	return module.LibraryFactory(config, module)
}

func getBinaryOrSharedLib(m blueprint.Module) (*Library, bool) {
	if sl, ok := m.(*SharedLibrary); ok {
		return &sl.Library, true
	} else if b, ok := m.(*Binary); ok {
		return &b.Library, true
	}

	return nil, false
}

// Traverse the dependency tree, following all StaticDepTag and WholeStaticDepTag links.
// Do *not* include modules which are in the tree via any other dependency tag.
func getLinkableModules(mctx blueprint.TopDownMutatorContext) map[blueprint.Module]bool {
	ret := make(map[blueprint.Module]bool)

	mctx.WalkDeps(func(dep blueprint.Module, parent blueprint.Module) bool {
		// Stop iteration once we get to other kinds of dependency which won't
		// actually be linked.
		if mctx.OtherModuleDependencyTag(dep) != StaticDepTag &&
			mctx.OtherModuleDependencyTag(dep) != WholeStaticDepTag {
			return false
		}
		ret[dep] = true

		return true
	})

	return ret
}

// Check that no libraries are being accidentally linked twice, by having one copy
// linked explicitly (via export_static_libs), and another included in a different
// library via whole_static_libs.
func checkForMultipleLinking(topLevelModuleName string, allExportStaticLibs map[string]bool, insideWholeLibs map[string]string) {
	duplicateDeps := []string{}
	for dep := range allExportStaticLibs {
		if _, ok := insideWholeLibs[dep]; ok {
			duplicateDeps = append(duplicateDeps, dep)
		}
	}

	if len(duplicateDeps) > 0 {
		msg := fmt.Sprintf("Warning: %s links with the following libraries multiple times:\n", topLevelModuleName)
		for _, dep := range duplicateDeps {
			msg += fmt.Sprintf("  * %s, but also %s, which includes %s as a whole_static_lib\n",
				dep, insideWholeLibs[dep], dep)
		}
		panic(msg)
	}
}

// While traversing the static library dependency tree, propagate extra properties.
func propagateOtherExportedProperties(l *Library, depLib *StaticLibrary) {
	props := l.Build()
	for _, shLib := range depLib.Properties.Export_shared_libs {
		if !utils.Contains(props.Shared_libs, shLib) {
			props.Shared_libs = append(props.Shared_libs, shLib)
			props.Extra_shared_libs = append(props.Extra_shared_libs, shLib)
		}
	}
	for _, ldlib := range depLib.Properties.Export_ldlibs {
		if !utils.Contains(props.Ldlibs, ldlib) {
			props.Ldlibs = append(props.Ldlibs, ldlib)
		}
	}
	props.Ldflags = append(props.Ldflags, depLib.Properties.Export_ldflags...)
}

func ExportLibFlagsMutator(mctx blueprint.TopDownMutatorContext) {
	l, ok := getBinaryOrSharedLib(mctx.Module())
	if !ok {
		return
	}

	// Track the set of everything mentioned in 'export_static_libs' of all
	// dependencies of this module, for multiple-link checking.
	allExportStaticLibs := make(map[string]bool)
	// Map between a library name and the first encountered lib in which it
	// is used in whole_static_libs.
	insideWholeLibs := make(map[string]string)
	// VisitDepsDepthFirst doesn't let us stop iteration, so get the list of
	// modules to examine separately using WalkDeps.
	modulesToVisit := getLinkableModules(mctx)

	mctx.VisitDepsDepthFirst(func(dep blueprint.Module) {
		if _, ok := modulesToVisit[dep]; !ok {
			return
		}

		if depLib, ok := dep.(*StaticLibrary); ok {
			for _, subLib := range depLib.Properties.Whole_static_libs {
				if firstContainingLib, ok := insideWholeLibs[subLib]; ok {
					panic(fmt.Sprintf("%s links with %s and %s, which both contain %s as whole_static_libs",
						mctx.Module().Name(), firstContainingLib,
						depLib.Name(), subLib))
				} else {
					insideWholeLibs[subLib] = depLib.Name()
				}
			}
			for _, subLib := range depLib.Properties.Export_static_libs {
				allExportStaticLibs[subLib] = true
			}

			propagateOtherExportedProperties(l, depLib)
		} else if _, ok := dep.(*GenerateStaticLibrary); ok {
			// Nothing to do for GeneratedStaticLibrary
			//
			// The GeneratedStaticLibrary is expected to be self
			// contained, so no pulling in of other static or shared
			// libraries.
		} else {
			panic(fmt.Sprintf("%s is not a StaticLibrary", dep.Name()))
		}

		// Don't add whole_static_lib components to the library list, because their
		// contents are already included in the parent library.
		if mctx.OtherModuleDependencyTag(dep) != WholeStaticDepTag && mctx.OtherModuleDependencyTag(dep) != StaticDepTag {
			panic(fmt.Sprintf("Non WholeStatic or Static dep tag encountered visiting %s from %s",
				dep.Name(), mctx.ModuleName()))
		}
	})

	checkForMultipleLinking(mctx.ModuleName(), allExportStaticLibs, insideWholeLibs)
}

type graphMutatorHandler struct {
	graph graph.Graph
}

const (
	maxInt = int(^uint(0) >> 1)
	minInt = -maxInt - 1
)

func (handler *graphMutatorHandler) ResolveDependencySortMutator(mctx blueprint.BottomUpMutatorContext) {
	mainModule := mctx.Module()
	if e, ok := mainModule.(Enableable); ok {
		if !IsEnabled(e) {
			return // Not enabled, so not needed
		}
	}
	if _, ok := mainModule.(*Defaults); ok {
		return // ignore bob_defaults
	}
	handler.graph.AddNode(mainModule.Name())

	var mainBuild *Build
	if buildProps, ok := mainModule.(ModuleWithBuildProps); ok {
		mainBuild = buildProps.Build()
	} else {
		return // ignore not a build
	}

	for _, lib := range mainBuild.Static_libs {
		if _, err := handler.graph.AddEdgeToExistingNodes(mainModule.Name(), lib); err != nil {
			panic(fmt.Errorf("'%s' depends on '%s', but '%s' is either not defined or disabled", mainModule.Name(), lib, lib))
		}
		handler.graph.SetEdgeColor(mainModule.Name(), lib, "blue")
	}

	for _, lib := range mainBuild.Export_static_libs {
		if _, err := handler.graph.AddEdgeToExistingNodes(mainModule.Name(), lib); err != nil {
			panic(fmt.Errorf("'%s' depends on '%s', but '%s' is either not defined or disabled", mainModule.Name(), lib, lib))
		}
		handler.graph.SetEdgeColor(mainModule.Name(), lib, "green")
	}

	for _, lib := range mainBuild.Whole_static_libs {
		if _, err := handler.graph.AddEdgeToExistingNodes(mainModule.Name(), lib); err != nil {
			panic(fmt.Errorf("'%s' depends on '%s', but '%s' is either not defined or disabled", mainModule.Name(), lib, lib))
		}
		handler.graph.SetEdgeColor(mainModule.Name(), lib, "red")
	}

	temporaryPaths := map[string][]string{} // For preserving order in declaration

	for i, previous := range mainBuild.Static_libs {
		for j := i + 1; j < len(mainBuild.Static_libs); j++ {
			lib := mainBuild.Static_libs[j]
			if !handler.graph.IsReachable(lib, previous) {
				if handler.graph.AddEdge(previous, lib) {
					temporaryPaths[previous] = append(temporaryPaths[previous], lib)
					handler.graph.SetEdgeColor(previous, lib, "pink")
				}
			}
		}
	}

	for i, previous := range mainBuild.Export_static_libs {
		for j := i + 1; j < len(mainBuild.Export_static_libs); j++ {
			lib := mainBuild.Export_static_libs[j]
			if !handler.graph.IsReachable(lib, previous) {
				if handler.graph.AddEdge(previous, lib) {
					temporaryPaths[previous] = append(temporaryPaths[previous], lib)
					handler.graph.SetEdgeColor(previous, lib, "pink")
				}
			}
		}
	}

	sub := graph.GetSubgraph(handler.graph, mainModule.Name())

	// Remove temporary path
	for key, list := range temporaryPaths {
		for _, value := range list {
			handler.graph.DeleteEdge(key, value)
		}
	}

	// The order of static libraries influences performance by
	// influencing memory layout. Where possible we want libraries
	// that depend on each other to be as close as possible. Library
	// order is determined by a topological sort.  Setting the
	// priority changes the order that child nodes are visited.
	//
	// Libraries that are frequently called are more
	// important and should be close to their callers. This information is not available in bob,
	// so estimate this with the number of users.
	//
	// Libraries that are large, or will cause a large number of
	// libraries to occur in the middle of the list, should be at the
	// end of the list. Treat this as the cost of visiting the
	// library. As an estimate of cost, count the number of libraries
	// that would be pulled in.
	//
	// The node priority is calculated as 'A * importance - cost',
	// where A is an arbitraty scaling factor.
	//
	// This is a bottom up mutator, so by the time we get to a binary
	// (or shared library), this mutator will have run on all their
	// dependencies and the (shared) graph will be complete (for the
	// current module).
	for _, nodeID := range sub.GetNodes() {
		cost := graph.GetSubgraph(sub, nodeID).GetNodeCount()
		sources, _ := sub.GetSources(nodeID)
		priority := len(sources)
		sub.SetNodePriority(nodeID, (10*priority)-cost)
	}

	// The main library must always be evaluated first in the topological sort
	sub.SetNodePriority(mainModule.Name(), minInt)

	// We want those edges for calculating priority. After setting priority we can remove them.
	sub.DeleteProxyEdges("red")

	sub2 := graph.GetSubgraph(sub, mainModule.Name())
	sortedStaticLibs, isDAG := graph.TopologicalSort(sub2)

	// Pop the module itself from the front of the list
	sortedStaticLibs = sortedStaticLibs[1:]

	if !isDAG {
		panic("We have detected cycle: " + mainModule.Name())
	} else {
		mainBuild.Resolved_static_libs = sortedStaticLibs
	}

	alreadyAddedStaticLibsDependencies := append(mainBuild.Static_libs, mainBuild.Export_static_libs...)
	extraStaticLibsDependencies := utils.Difference(mainBuild.Resolved_static_libs, alreadyAddedStaticLibsDependencies)

	mctx.AddVariationDependencies(nil, StaticDepTag, extraStaticLibsDependencies...)

	// This module may now depend on extra shared libraries, inherited from included
	// static libraries. Add that dependency here.
	mctx.AddVariationDependencies(nil, SharedDepTag, mainBuild.Extra_shared_libs...)
}
