/*
 * Copyright:
 * ----------------------------------------------------------------------------
 * This confidential and proprietary software may be used only as authorized
 * by a licensing agreement from ARM Limited.
 *      (C) COPYRIGHT 2016-2018 ARM Limited, ALL RIGHTS RESERVED
 * The entire notice above must be reproduced on all authorized copies and
 * copies may only be made to the extent permitted by a licensing agreement
 * from ARM Limited.
 * ----------------------------------------------------------------------------
 */

package core

import (
	"errors"
	"os"
	"path"
	"path/filepath"
	"runtime"
	"strconv"
	"strings"

	"github.com/google/blueprint"

	"bob/utils"
)

var (
	builddir_var = pctx.StaticVariable("BuildDir", builddir)
	srcdir_var   = pctx.StaticVariable("SrcDir", srcdir)
	pctx         = blueprint.NewPackageContext("bob")
)

type LinuxGenerator struct{}

// Convert a path to a library into a compiler flag.
// This needs to strip any path, file extension, lib prefix, and prepend -l
func pathToLibFlag(path string) string {
	_, base := filepath.Split(path)
	ext := filepath.Ext(base)
	base = strings.TrimSuffix(base, ext)
	base = strings.TrimPrefix(base, "lib")
	return "-l" + base
}

func AddPhony(p PhonyInterface, ctx blueprint.ModuleContext,
	installDeps []string, optional bool) {

	deps := append(p.Outputs(getBackend(ctx)), installDeps...)

	ctx.Build(pctx,
		blueprint.BuildParams{
			Rule:     blueprint.Phony,
			Inputs:   deps,
			Outputs:  []string{p.ShortName()},
			Optional: optional,
		})
}

func (g *LinuxGenerator) sourcePrefix() string {
	return "$SrcDir"
}

func (g *LinuxGenerator) buildDir() string {
	return "$BuildDir"
}

func (g *LinuxGenerator) sourceOutputDir(m *GenerateCommon) string {
	return filepath.Join("$BuildDir", "gen", m.Name())
}

var copyRule = pctx.StaticRule("copy",
	blueprint.RuleParams{
		Command:     "cp --reflink=auto $in $out",
		Description: "$out",
	})

type targetableModule interface {
	blueprint.Module
	getTarget() string
}

// Where to put generated shared libraries to simplify linking
// As long as the module is targetable, we can infer the library path
func getSharedLibLinkPath(t targetableModule) string {
	return filepath.Join("$BuildDir", t.getTarget(), "shared", t.Name()+".so")
}

// Where to put generated binaries in order to make sure generated binaries
// are available in the same directory as compiled binaries
func getBinaryPath(t targetableModule) string {
	return filepath.Join("$BuildDir", t.getTarget(), "executable", t.Name())
}

// Generate the build actions for a GenerateSource module and populates the outputs.
func (g *LinuxGenerator) GenerateCommonActions(m *GenerateCommon, ctx blueprint.ModuleContext, inouts []Inout) {

	cmd, args, implicits, host_target := m.GetArgs(ctx)

	ld_library_path := ""
	if _, ok := args["host_bin"]; ok {
		ld_library_path += "LD_LIBRARY_PATH=" + filepath.Join("${BuildDir}", host_target, "shared") + ":$$LD_LIBRARY_PATH "
	}
	utils.StripUnusedArgs(args, cmd)
	args["depfile"] = ""

	var pool blueprint.Pool
	if m.Properties.Console {
		// Console can be used to run longrunning jobs (even interactive jobs).
		pool = blueprint.Console
	}

	rulename := "gen_" + m.Name()
	//print("Keys:" + strings.Join(argkeys, ",") + "\n")
	rule := ctx.Rule(pctx,
		rulename,
		blueprint.RuleParams{
			Command: ld_library_path + cmd,
			// Restat is always set to true. This is due to wanting to enable scripts
			// to only update the outputs if they have changed (keeping the same mtime if it
			// has not). If there are no updates, the following rules will not have to update
			// the output.
			Restat:      true,
			Pool:        pool,
			Depfile:     "$depfile",
			Description: "$out",
		},
		utils.SortedKeys(args)...)

	for _, inout := range inouts {
		if _, ok := args["headers_generated"]; ok {
			headers := utils.Filter(inout.out, utils.IsHeader)
			args["headers_generated"] = strings.Join(headers, " ")
		}
		if _, ok := args["srcs_generated"]; ok {
			sources := utils.Filter(inout.out, utils.IsSource)
			args["srcs_generated"] = strings.Join(sources, " ")
		}

		args["depfile"] = inout.depfile
		implicits = append(implicits, inout.implicit_srcs...)

		ctx.Build(pctx,
			blueprint.BuildParams{
				Rule:      rule,
				Inputs:    append(inout.src_in, inout.gen_in...),
				Outputs:   inout.out,
				Implicits: implicits,
				Args:      args,
				Optional:  true,
			})
	}
}

func (g *LinuxGenerator) GenerateSourceActions(m *GenerateSource, ctx blueprint.ModuleContext, inouts []Inout) {
	g.GenerateCommonActions(&m.GenerateCommon, ctx, inouts)

	installDeps := g.install(m, ctx)
	AddPhony(m, ctx, installDeps, !IsBuiltByDefault(m))
}

func (g *LinuxGenerator) TransformSourceActions(m *TransformSource, ctx blueprint.ModuleContext, inouts []Inout) {
	g.GenerateCommonActions(&m.GenerateCommon, ctx, inouts)

	installDeps := g.install(m, ctx)
	AddPhony(m, ctx, installDeps, !IsBuiltByDefault(m))
}

func (g *LinuxGenerator) GenStaticActions(m *GenerateStaticLibrary, ctx blueprint.ModuleContext, inouts []Inout) {
	g.GenerateCommonActions(&m.GenerateCommon, ctx, inouts)

	installDeps := g.install(m, ctx)
	AddPhony(m, ctx, installDeps, !IsBuiltByDefault(m))
}

func (g *LinuxGenerator) GenSharedActions(m *GenerateSharedLibrary, ctx blueprint.ModuleContext, inouts []Inout) {
	g.GenerateCommonActions(&m.GenerateCommon, ctx, inouts)

	// Create a rule to copy the generated library
	// from gen_dir to the common library directory
	ctx.Build(pctx,
		blueprint.BuildParams{
			Rule:     copyRule,
			Inputs:   []string{getLibraryGeneratedPath(m, g)},
			Outputs:  []string{getSharedLibLinkPath(m)},
			Optional: true,
		})

	installDeps := g.install(m, ctx)
	AddPhony(m, ctx, installDeps, !IsBuiltByDefault(m))
}

func (g *LinuxGenerator) GenBinaryActions(m *GenerateBinary, ctx blueprint.ModuleContext, inouts []Inout) {
	g.GenerateCommonActions(&m.GenerateCommon, ctx, inouts)

	// Create a rule to copy the generated binary
	// from gen_dir to the common binary directory
	ctx.Build(pctx,
		blueprint.BuildParams{
			Rule:     copyRule,
			Inputs:   []string{getLibraryGeneratedPath(m, g)},
			Outputs:  []string{getBinaryPath(m)},
			Optional: true,
		})

	installDeps := g.install(m, ctx)
	AddPhony(m, ctx, installDeps, !IsBuiltByDefault(m))
}

var asRule = pctx.StaticRule("as",
	blueprint.RuleParams{
		Depfile:     "$out.d",
		Deps:        blueprint.DepsGCC,
		Command:     "$build_wrapper $ascompiler $asflags $in -MD $out.d -o $out",
		Description: "$out",
	}, "ascompiler", "asflags", "build_wrapper")

var ccRule = pctx.StaticRule("cc",
	blueprint.RuleParams{
		Depfile:     "$out.d",
		Deps:        blueprint.DepsGCC,
		Command:     "$build_wrapper $ccompiler -c $cflags $conlyflags -MMD -MF $out.d $in -o $out",
		Description: "$out",
	}, "ccompiler", "cflags", "conlyflags", "build_wrapper")

var cxxRule = pctx.StaticRule("cxx",
	blueprint.RuleParams{
		Depfile:     "$out.d",
		Deps:        blueprint.DepsGCC,
		Command:     "$build_wrapper $cxxcompiler -c $cflags $cxxflags -MMD -MF $out.d $in -o $out",
		Description: "$out",
	}, "cxxcompiler", "cflags", "cxxflags", "build_wrapper")

func (l *Library) ObjDir() string {
	return filepath.Join("${BuildDir}", l.Properties.Target_type, "objects", l.Name()) + string(os.PathSeparator)
}

// This function has common support to compile objs for static libs, shared libs and binaries.
func (l *Library) CompileObjs(ctx blueprint.ModuleContext) []string {
	g := getBackend(ctx)
	srcs := l.GetSrcs(ctx)

	exported_includes, exported_cflags := l.GetExportedVariables(ctx)
	// There are 2 sets of include dirs - "global" and "local".
	// Local acts on the root source directory.

	// The order we want is  local_include_dirs, export_local_include_dirs,
	//                       include_dirs, export_include_dirs
	include_dirs := l.Properties.Local_include_dirs
	include_dirs = append(include_dirs, l.Properties.Export_local_include_dirs...)
	include_dirs = append(include_dirs, l.Properties.Include_dirs...)
	include_dirs = append(include_dirs, l.Properties.Export_include_dirs...)
	include_dirs = append(include_dirs, exported_includes...)

	gendirs, order_only := l.GetGeneratedHeaders(ctx)
	include_dirs = append(include_dirs, gendirs...)
	include_flags := utils.PrefixAll(include_dirs, "-I")
	cflags_list := append(l.Properties.Cflags, include_flags...)
	cflags_list = append(cflags_list, l.Properties.Export_cflags...)
	cflags_list = append(cflags_list, exported_cflags...)

	as, astargetflags := GetAssembler(ctx, l.Properties.Target_type)
	cc, cctargetflags := GetCCompiler(ctx, l.Properties.Target_type)
	cxx, cxxtargetflags := GetCXXCompiler(ctx, l.Properties.Target_type)

	ctx.Variable(pctx, "asflags", astargetflags+" "+strings.Join(l.Properties.Asflags, " "))
	ctx.Variable(pctx, "cflags", strings.Join(cflags_list, " "))
	ctx.Variable(pctx, "conlyflags", cctargetflags+" "+strings.Join(l.Properties.Conlyflags, " "))
	ctx.Variable(pctx, "cxxflags", cxxtargetflags+" "+strings.Join(l.Properties.Cxxflags, " "))

	var object_files []string
	for _, source := range srcs {
		var rule blueprint.Rule
		args := make(map[string]string)
		switch path.Ext(source) {
		case ".s":
			args["ascompiler"] = as
			args["asflags"] = "$asflags"
			rule = asRule
		case ".S":
			// Assembly with .S suffix must be preprocessed by the C compiler
			fallthrough
		case ".c":
			args["ccompiler"] = cc
			args["cflags"] = "$cflags"
			args["conlyflags"] = "$conlyflags"
			rule = ccRule
		case ".cc":
			fallthrough
		case ".cpp":
			args["cxxcompiler"] = cxx
			args["cflags"] = "$cflags"
			args["cxxflags"] = "$cxxflags"
			rule = cxxRule
		default:
			panic(errors.New("Files with extension '" + path.Ext(source) + "' not supported"))
		}

		buildWrapper, buildWrapperDeps := l.Properties.Build.getBuildWrapperAndDeps(ctx)
		args["build_wrapper"] = buildWrapper

		sourceWithoutBuildDir := source
		if buildDir := g.buildDir(); strings.HasPrefix(sourceWithoutBuildDir, buildDir) {
			sourceWithoutBuildDir = sourceWithoutBuildDir[len(buildDir):]
		}
		var output = l.ObjDir() + sourceWithoutBuildDir + ".o"

		ctx.Build(pctx,
			blueprint.BuildParams{
				Rule:      rule,
				Outputs:   []string{output},
				Inputs:    []string{source},
				Args:      args,
				OrderOnly: append(order_only, buildWrapperDeps...),
				Optional:  true,
			})
		object_files = append(object_files, output)
	}

	return object_files
}

// Returns all the source files for a C/C++ library. This includes any sources that are generated.
func (l *Library) GetSrcs(ctx blueprint.ModuleContext) []string {
	g := getBackend(ctx)
	srcs := l.Properties.GetSrcs(ctx)
	srcs = append(srcs, l.Properties.Build.SourceProps.Specials...)
	ctx.VisitDirectDepsIf(
		func(m blueprint.Module) bool { return ctx.OtherModuleDependencyTag(m) == generatedSourceTag },
		func(m blueprint.Module) {
			if gs, ok := m.(DependentInterface); ok {
				srcs = append(srcs, getSourcesGenerated(g, gs)...)
			} else {
				panic(errors.New(ctx.OtherModuleName(m) + " does not have outputs"))
			}
		})
	return srcs
}

// Returns the whole static dependencies for a library.
// Note that this is transitive, so it includes both the static library and all
// its whole static dependencies.
// While this is a break with the design idea of not having transitive dependencies,
// this is done due to the fact that merging two static libraries are not that easy.
func (l *Library) GetWholeStaticLibs(ctx blueprint.ModuleContext) []string {
	g := getBackend(ctx)
	libs := []string{}
	ctx.VisitDirectDepsIf(
		func(m blueprint.Module) bool { return ctx.OtherModuleDependencyTag(m) == WholeStaticDepTag },
		func(m blueprint.Module) {
			if sl, ok := m.(*StaticLibrary); ok {
				libs = append(libs, sl.Outputs(g)...)
			} else if sl, ok := m.(*GenerateStaticLibrary); ok {
				libs = append(libs, getLibraryGeneratedPath(sl, g))
			} else {
				panic(errors.New(ctx.OtherModuleName(m) + " is not a static library"))
			}
		})

	return libs
}

// Returns all the static library dependencies for a module.
func (l *Library) GetStaticLibs(ctx blueprint.ModuleContext) []string {
	g := getBackend(ctx)
	libs := []string{}
	for _, moduleName := range l.Properties.Resolved_static_libs {
		dep, _ := ctx.GetDirectDep(moduleName)
		if sl, ok := dep.(*StaticLibrary); ok {
			libs = append(libs, sl.Outputs(g)...)
		} else if sl, ok := dep.(*GenerateStaticLibrary); ok {
			libs = append(libs, getLibraryGeneratedPath(sl, g))
		} else {
			panic(errors.New(ctx.OtherModuleName(dep) + " is not a static library"))
		}
	}

	return libs
}

func (g *LinuxGenerator) staticLibOutputDir(m *StaticLibrary) string {
	return filepath.Join("${BuildDir}", m.Properties.Target_type, "static")
}

// The rule for building a static library
// Note that we need to remove the old library, else we will not remove the old object files
var staticLibraryRule = pctx.StaticRule("static_library",
	blueprint.RuleParams{
		Command:     "rm -f $out && $build_wrapper $ar -rcs $out $in",
		Description: "$out",
	}, "ar", "build_wrapper")

var wholeStaticLibraryRule = pctx.StaticRule("whole_static_library",
	blueprint.RuleParams{
		Command: "rm -f $out && { " +
			"echo create $out; " +
			"for i in $in; do echo addmod $$i; done; " +
			"for i in $whole_static_libs; do echo addlib $$i; done; " +
			"echo save; echo end; } " +
			"| $build_wrapper $ar -M && test -f $out",
		Description: "$out",
	}, "ar", "build_wrapper", "whole_static_libs")

func (g *LinuxGenerator) StaticActions(m *StaticLibrary, ctx blueprint.ModuleContext) {
	if len(m.Properties.Static_libs) > 0 || len(m.Properties.Shared_libs) > 0 {
		panic(errors.New("static library cannot include another library, only include it using whole_static"))
	}

	rule := staticLibraryRule

	buildWrapper, buildWrapperDeps := m.Properties.Build.getBuildWrapperAndDeps(ctx)

	args := map[string]string{
		"ar":            GetArchiver(ctx, m.Properties.Target_type),
		"build_wrapper": buildWrapper,
	}

	wholeStaticLibs := m.Library.GetWholeStaticLibs(ctx)

	if len(wholeStaticLibs) > 0 {
		rule = wholeStaticLibraryRule
		args["whole_static_libs"] = strings.Join(wholeStaticLibs, " ")
	}

	ctx.Build(pctx,
		blueprint.BuildParams{
			Rule:      rule,
			Outputs:   m.Outputs(g),
			Inputs:    m.Library.CompileObjs(ctx),
			Implicits: wholeStaticLibs,
			OrderOnly: buildWrapperDeps,
			Optional:  true,
			Args:      args,
		})

	installDeps := g.install(m, ctx)
	AddPhony(m, ctx, installDeps, !IsBuiltByDefault(m))
}

// This section contains functions that are common for shared libraries and executables.

func (l *Library) getSharedLibLinkPaths(ctx blueprint.ModuleContext) (libs []string) {
	ctx.VisitDirectDepsIf(
		func(m blueprint.Module) bool { return ctx.OtherModuleDependencyTag(m) == SharedDepTag },
		func(m blueprint.Module) {
			if t, ok := m.(targetableModule); ok {
				libs = append(libs, getSharedLibLinkPath(t))

			} else {
				panic(errors.New(ctx.OtherModuleName(m) + " doesn't support targets"))
			}
		})
	return
}

func (l *Library) getSharedLibFlags(ctx blueprint.ModuleContext) (flags []string) {
	// With forwarding shared library we do not have to use --no-as-needed for dependencies because it is already set
	use_no_as_needed := !l.Build().IsForwardingSharedLibrary()

	ctx.VisitDirectDepsIf(
		func(m blueprint.Module) bool { return ctx.OtherModuleDependencyTag(m) == SharedDepTag },
		func(m blueprint.Module) {
			if sl, ok := m.(*SharedLibrary); ok {
				b := sl.Build()
				if b.IsForwardingSharedLibrary() {
					flags = append(flags, "-Wl,--copy-dt-needed-entries")
					if use_no_as_needed {
						flags = append(flags, "-Wl,--no-as-needed")
					}
				}
				flags = append(flags, pathToLibFlag(sl.Name()))
				if b.IsForwardingSharedLibrary() {
					if use_no_as_needed {
						flags = append(flags, "-Wl,--as-needed")
					}
					flags = append(flags, "-Wl,--no-copy-dt-needed-entries")
				}
			} else if sl, ok := m.(*GenerateSharedLibrary); ok {
				flags = append(flags, pathToLibFlag(sl.Name()))
			} else {
				panic(errors.New(ctx.OtherModuleName(m) + " is not a shared library"))
			}
		})
	return
}

func (l *Library) getSharedLibraryDir() string {
	return filepath.Join("${BuildDir}", l.Properties.Target_type, "shared")
}

func (g *LinuxGenerator) sharedLibOutputDir(m *SharedLibrary) string {
	return m.Library.getSharedLibraryDir()
}

func (g *LinuxGenerator) sharedLibsDir(targetType string) string {
	return filepath.Join("${BuildDir}", targetType, "shared")
}

func (l *Library) getCommonLibArgs(ctx blueprint.ModuleContext) map[string]string {
	ldflags := l.Properties.Ldflags

	if l.Build().IsForwardingSharedLibrary() {
		ldflags = append(ldflags, "-Wl,--no-as-needed")
	} else {
		ldflags = append(ldflags, "-Wl,--as-needed")
	}

	shared_lib_flags := l.getSharedLibFlags(ctx)

	compiler, cctargetflags := GetCXXCompiler(ctx, l.Properties.Target_type)
	build_wrapper, _ := l.Properties.Build.getBuildWrapperAndDeps(ctx)

	args := map[string]string{
		"build_wrapper":     build_wrapper,
		"compiler":          compiler,
		"ldflags":           cctargetflags + " " + strings.Join(ldflags, " "),
		"shared_libs_dir":   l.getSharedLibraryDir(),
		"shared_libs_flags": strings.Join(shared_lib_flags, " "),
		"static_libs":       strings.Join(l.GetStaticLibs(ctx), " "),
		"ldlibs":            strings.Join(l.Properties.Ldlibs, " "),
		"whole_static_libs": strings.Join(l.GetWholeStaticLibs(ctx), " "),
	}
	return args
}

func (l *SharedLibrary) getLibArgs(ctx blueprint.ModuleContext) map[string]string {
	args := l.getCommonLibArgs(ctx)
	ldflags := []string{}

	if l.Properties.Library_version != "" {
		var soname_flag = "-Wl,-soname," + l.getSoname()
		ldflags = append(ldflags, soname_flag)
	}

	args["ldflags"] += " " + strings.Join(ldflags, " ")

	return args
}

func (b *Binary) getLibArgs(ctx blueprint.ModuleContext) map[string]string {
	return b.getCommonLibArgs(ctx)
}

// Returns the implicit dependencies for a library
func (l *Library) Implicits(ctx blueprint.ModuleContext) []string {
	implicits := append(l.GetWholeStaticLibs(ctx), l.GetStaticLibs(ctx)...)
	implicits = append(implicits, l.getSharedLibLinkPaths(ctx)...)

	return implicits
}

// Return the parallelism that the current build environment can support
func getParallelism() int {
	// This environment variable is used to indicate available
	// parallelism in a certain build farm. It needs to be used in
	// preference to the detected number of CPUs
	if str, ok := os.LookupEnv("MPDTI_BUILD_PARALLELISM"); ok {
		if p, err := strconv.Atoi(str); err == nil {
			return p
		}
	}
	return runtime.NumCPU()
}

var parallelism = getParallelism()

var linkPoolParams = blueprint.PoolParams{
	Comment: "Limit the parallelization of linking, which is memory intensive",
	Depth:   (getParallelism() / 5) + 1,
}

var linkPool = pctx.StaticPool("link", linkPoolParams)

var sharedLibraryRule = pctx.StaticRule("shared_library",
	blueprint.RuleParams{
		Command: "$build_wrapper $compiler -shared $in -o $out $ldflags " +
			"-Wl,--whole-archive  $whole_static_libs -Wl,--no-whole-archive $static_libs " +
			"-Wl,-rpath-link,$shared_libs_dir -L$shared_libs_dir $shared_libs_flags $ldlibs",
		Description: "$out",
		Pool:        linkPool,
	}, "build_wrapper", "compiler", "ldflags", "ldlibs", "shared_libs_dir", "shared_libs_flags",
	"static_libs", "whole_static_libs")

var symlinkRule = pctx.StaticRule("symlink",
	blueprint.RuleParams{
		Command:     "for i in $out; do ln -nsf $target $$i; done;",
		Description: "$out",
	}, "target")

func (g *LinuxGenerator) SharedActions(m *SharedLibrary, ctx blueprint.ModuleContext) {
	object_files := m.CompileObjs(ctx)

	_, buildWrapperDeps := m.Properties.Build.getBuildWrapperAndDeps(ctx)

	installDeps := g.install(m, ctx)

	// Create symlinks if needed
	for name, symlink_tgt := range m.LibrarySymlinks(ctx) {
		symlink := filepath.Join(m.OutputDir(g), name)
		lib := filepath.Join(m.OutputDir(g), symlink_tgt)
		ctx.Build(pctx,
			blueprint.BuildParams{
				Rule:     symlinkRule,
				Inputs:   []string{lib},
				Outputs:  []string{symlink},
				Args:     map[string]string{"target": symlink_tgt},
				Optional: true,
			})
		installDeps = append(installDeps, symlink)
	}

	ctx.Build(pctx,
		blueprint.BuildParams{
			Rule:      sharedLibraryRule,
			Outputs:   m.Outputs(g),
			Inputs:    object_files,
			Implicits: m.Library.Implicits(ctx),
			OrderOnly: buildWrapperDeps,
			Optional:  true,
			Args:      m.getLibArgs(ctx),
		})

	AddPhony(m, ctx, installDeps, !IsBuiltByDefault(m))
}

func (g *LinuxGenerator) binaryOutputDir(m *Binary) string {
	return filepath.Join("${BuildDir}", m.Properties.Target_type, "executable")
}

var executableRule = pctx.StaticRule("executable",
	blueprint.RuleParams{
		Command: "$build_wrapper $compiler $in -o $out $ldflags $static_libs " +
			"-Wl,-rpath-link,$shared_libs_dir -L$shared_libs_dir $shared_libs_flags $ldlibs",
		Description: "$out",
		Pool:        linkPool,
	}, "build_wrapper", "compiler", "ldflags", "ldlibs", "shared_libs_dir", "shared_libs_flags",
	"static_libs", "whole_static_libs")

func (g *LinuxGenerator) BinaryActions(m *Binary, ctx blueprint.ModuleContext) {
	object_files := m.CompileObjs(ctx)
	/* By default, build all target binaries */
	optional := !IsBuiltByDefault(m)

	_, buildWrapperDeps := m.Properties.Build.getBuildWrapperAndDeps(ctx)

	ctx.Build(pctx,
		blueprint.BuildParams{
			Rule:      executableRule,
			Outputs:   m.Outputs(g),
			Inputs:    object_files,
			Implicits: m.Library.Implicits(ctx),
			OrderOnly: buildWrapperDeps,
			Optional:  true,
			Args:      m.getLibArgs(ctx),
		})
	installDeps := g.install(m, ctx)
	AddPhony(m, ctx, installDeps, optional)
}

func (*LinuxGenerator) AliasActions(m *Alias, ctx blueprint.ModuleContext) {
	srcs := []string{}

	/* Only depend on enabled targets */
	ctx.VisitDirectDepsIf(
		func(p blueprint.Module) bool { return ctx.OtherModuleDependencyTag(p) == aliasTag },
		func(p blueprint.Module) {
			if e, ok := p.(Enableable); ok {
				if !IsEnabled(e) {
					return
				}
			}
			srcs = append(srcs, ctx.OtherModuleName(p))
		})

	ctx.Build(pctx,
		blueprint.BuildParams{
			Rule:     blueprint.Phony,
			Inputs:   srcs,
			Outputs:  []string{m.Name()},
			Optional: true,
		})
}

var installRule = pctx.StaticRule("install",
	blueprint.RuleParams{
		Command:     "cp --remove-destination --reflink=auto $in $out",
		Description: "$out",
	})

func (*LinuxGenerator) install(m interface{}, ctx blueprint.ModuleContext) []string {
	ins := m.(Installable)

	installPath, ok := getInstallGroupPath(ctx)
	if !ok {
		return []string{}
	}
	installPath = filepath.Join("${BuildDir}", installPath)

	props := ins.GetInstallableProps()
	if props.Relative_install_path != "" {
		installPath = filepath.Join(installPath, props.Relative_install_path)
	}

	installedFiles := []string{}

	rule := installRule
	args := map[string]string{}
	deps := []string{}
	if props.Post_install_cmd != "" {
		rulename := "install"
		tool := filepath.Join(ctx.ModuleDir(), props.Post_install_tool)

		// The command may or may not modify the output
		// --reflink=auto should work in either case.
		cmd := "cp --reflink=auto $in $out ; " + props.Post_install_cmd

		args["bob_config"] = configPath
		args["tool"] = tool
		utils.StripUnusedArgs(args, props.Post_install_cmd)

		rule = ctx.Rule(pctx,
			rulename,
			blueprint.RuleParams{
				Command:     cmd,
				Description: "$out",
			},
			utils.SortedKeys(args)...)
		deps = append(deps, tool)
	}

	for _, src := range ins.FilesToInstall(ctx) {
		dest := filepath.Join(installPath, filepath.Base(src))

		ctx.Build(pctx,
			blueprint.BuildParams{
				Rule:      rule,
				Outputs:   []string{dest},
				Inputs:    []string{src},
				Args:      args,
				Implicits: deps,
				Optional:  true,
			})

		installedFiles = append(installedFiles, dest)
	}

	if symlink_ins, ok := m.(SymlinkInstaller); ok {
		symlinks := symlink_ins.LibrarySymlinks(ctx)

		for key, value := range symlinks {
			symlink := filepath.Join(installPath, key)
			symlink_tgt := filepath.Join(installPath, value)
			ctx.Build(pctx,
				blueprint.BuildParams{
					Rule:     symlinkRule,
					Outputs:  []string{symlink},
					Inputs:   []string{symlink_tgt},
					Args:     map[string]string{"target": value},
					Optional: true,
				})

			installedFiles = append(installedFiles, symlink)
		}
	}

	return append(installedFiles, ins.GetInstallDepPhonyNames(ctx)...)
}

func (g *LinuxGenerator) ResourceActions(m *Resource, ctx blueprint.ModuleContext) {
	installDeps := g.install(m, ctx)
	AddPhony(m, ctx, installDeps, false)
}

var copyIfExistsRule = pctx.StaticRule("cp_if_exists",
	blueprint.RuleParams{
		Command: "test -f $in && cp $in $out || (>&2 echo '" +
			"Warning: $in was not built - it has most likely " +
			"been disabled by your kernel config.' )",
		Description: "$out",
	})

func copyFileIfExists(ctx blueprint.ModuleContext, source string, dest string) {
	ctx.Build(pctx,
		blueprint.BuildParams{
			Rule:     copyIfExistsRule,
			Outputs:  []string{dest},
			Inputs:   []string{source},
			Optional: true,
		})
}

func (g *LinuxGenerator) kernelModOutputDir(m *KernelModule) string {
	return filepath.Join("${BuildDir}", "target", "kernel_modules", m.Name())
}

var kbuildRule = pctx.StaticRule("kbuild",
	blueprint.RuleParams{
		Command: "python $kmod_build -o $out --depfile $out.d " +
			"--module-dir $output_module_dir $extra_includes " +
			"--sources $in $kbuild_extra_symbols " +
			"--kernel $kernel_dir --cross-compile '$kernel_compiler' " +
			"$kbuild_options --extra-cflags '$extra_cflags' $make_args",
		Depfile:     "$out.d",
		Deps:        blueprint.DepsGCC,
		Pool:        blueprint.Console,
		Description: "$out",
	}, "kmod_build", "extra_includes", "extra_cflags", "kbuild_extra_symbols", "kernel_dir",
	"kernel_compiler", "kbuild_options", "make_args", "output_module_dir")

func (g *LinuxGenerator) KernelModuleActions(m *KernelModule, ctx blueprint.ModuleContext) {
	builtModule := filepath.Join(g.kernelModOutputDir(m), m.Name()+".ko")

	args := m.generateKbuildArgs(ctx)

	ctx.Build(pctx,
		blueprint.BuildParams{
			Rule:      kbuildRule,
			Outputs:   []string{builtModule},
			Inputs:    append(m.Properties.GetSrcs(ctx), m.Properties.Build.SourceProps.Specials...),
			Implicits: append(m.extraSymbolsFiles(ctx), args["copy_with_deps"]),
			Optional:  false,
			Args:      args,
		})

	// Add a dependency between Module.symvers and the kernel module. This
	// should really be added to Outputs or ImplicitOutputs above, but
	// Ninja doesn't support dependency files with multiple outputs yet.
	ctx.Build(pctx,
		blueprint.BuildParams{
			Rule:     blueprint.Phony,
			Inputs:   []string{builtModule},
			Outputs:  []string{filepath.Join(g.kernelModOutputDir(m), "Module.symvers")},
			Optional: true,
		})

	installDeps := g.install(m, ctx)
	AddPhony(m, ctx, installDeps, false)
}

func (*LinuxGenerator) Init(*blueprint.Context) {
}
