# Copyright:
# ----------------------------------------------------------------------------
# This confidential and proprietary software may be used only as authorized
# by a licensing agreement from ARM Limited.
#      (C) COPYRIGHT 2016-2018 ARM Limited, ALL RIGHTS RESERVED
# The entire notice above must be reproduced on all authorized copies and
# copies may only be made to the extent permitted by a licensing agreement
# from ARM Limited.
# ----------------------------------------------------------------------------
#

import os
import sys
import logging
import subprocess
import re
import tempfile

from config_system import get_config, set_config

logger = logging.getLogger(__name__)

def which_binary(executable, path=None):
    '''
    Emulates 'which', trying to locate a binary 'executable' in the 'path',
    or environment '$PATH', if none is given.
    '''

    if not executable.startswith('/'):
        if path is None:
            path = os.environ['PATH']
        paths = path.split(os.pathsep)
        for p in paths:
            f = os.path.join(p, executable)
            if os.path.isfile(f) and os.access(f, os.X_OK):
                return f
        logger.error("Executable (%s) not found in path (%s)", executable, path)
    else:
        if os.path.isfile(executable) and os.access(executable, os.X_OK):
            return executable
        logger.error("Executable (%s) not found", executable)

    return None

def check_output(command, dir=None):
    '''
    Executes the command, while making sure the executable is found in the $PATH,
    and returns the output. If the executable wasn't found, returns an empty string.
    The 'command' needs to be an array of arguments.
    '''
    executable = which_binary(command[0])
    output = ''
    if executable:
        try:
            output = subprocess.check_output(command, cwd=dir).strip()
            output = output.decode(sys.getdefaultencoding())
        except (OSError, subprocess.CalledProcessError) as e:
            logger.warning("Problem executing command: %s" % str(e))

    return output

def compiler_config():
    '''
    Adds runtime-dependant entries by analyzing user-configured flags.
    This is mainly required to allow some configuration-specific flags that are
    impossible or hard to express in Blueprint, because of the need to, say, run
    a binary and get the output from it.
    '''
    extra_target_ldflags = ''
    extra_host_ldflags = ''
    host_cxx = ''
    # No toolchain prefix indicates a native build
    native_build = get_config('TOOLCHAIN_PREFIX')['value'] == ''
    if get_config('TOOLCHAIN_CLANG')['value'] == 'y':
        host_cxx = get_config('CLANGXX_BINARY')['value']
        host_libstdcxx_path = check_output([host_cxx, '-print-file-name=libstdc++.so'])
        if native_build:
            target_libstdcxx_path = host_libstdcxx_path
        else:
            cross_gcc = which_binary(get_config('TOOLCHAIN_PREFIX')['value'] +
                get_config('GCC_BINARY')['value'])
            flags = get_config('GCC_TARGET_FLAGS')['value'].split(" ")
            flags = list(filter(None, flags))

            cross_sysroot = check_output([cross_gcc] + flags + ['-print-sysroot'])
            set_config('CLANG_SYSROOT', cross_sysroot)
            cross_version = check_output([cross_gcc] + flags + ['-dumpversion'])
            set_config('TARGET_TOOLCHAIN_VERSION', cross_version)
            target_libstdcxx_path = check_output([cross_gcc] + flags + ['-print-file-name=libstdc++.so'])
            crt_path = os.path.split(check_output([cross_gcc] + flags + ['-print-file-name=crt1.o']))[0]
            if crt_path != '':
                extra_target_ldflags += '-B{0} '.format(crt_path)

            libgcc_static_path = os.path.split(check_output([cross_gcc] + flags + ['-print-file-name=libgcc.a']))[0]
            if libgcc_static_path != '':
                extra_target_ldflags += '-B{0} -L{0} '.format(libgcc_static_path)

            crosslib_path = os.path.split(check_output([cross_gcc] + flags + ['-print-file-name=libgcc_s.so']))[0]
            if crosslib_path != '':
                extra_target_ldflags += '-L{0} '.format(crosslib_path)

    elif get_config('TOOLCHAIN_GNU')['value'] == 'y':
        host_cxx = get_config('GXX_BINARY')['value']
        flags = get_config('GCC_TARGET_FLAGS')['value'].split(" ")
        flags = list(filter(None, flags))
        host_libstdcxx_path = check_output([host_cxx,'-print-file-name=libstdc++.so'])
        target_libstdcxx_path = check_output([host_cxx] + flags + ['-print-file-name=libstdc++.so'])

    host_libstdcxx_dir = os.path.split(host_libstdcxx_path)[0]
    ld_library_path = os.getenv('LD_LIBRARY_PATH','')
    ld_paths = ld_library_path.split(":")
    if len(host_libstdcxx_dir) > 0:
        if host_libstdcxx_dir not in ld_paths:
            extra_host_ldflags += '-Wl,--enable-new-dtags -Wl,-rpath,{0} '.format(host_libstdcxx_dir)
            set_config('EXTRA_LD_LIBRARY_PATH', host_libstdcxx_dir)

    # In native builds we require the addition of rpath to the executable if not in load library path
    if native_build:
        target_libstdcxx_dir = os.path.split(target_libstdcxx_path)[0]
        if len(target_libstdcxx_dir) > 0:
            if target_libstdcxx_dir not in ld_paths:
                extra_target_ldflags += '-Wl,--enable-new-dtags -Wl,-rpath,{0} '.format(target_libstdcxx_dir)

    builder_android = get_config('BUILDER_ANDROID')['value'] == 'y'
    if not builder_android:
        cross_ld = which_binary(get_config('TOOLCHAIN_PREFIX')['value'] + 'ld')
        if cross_ld:
            cross_ld_version = check_output([cross_ld, '-version'])
            if cross_ld_version.count('gold') == 1:
                extra_target_ldflags += '-fuse-ld=bfd '

        host_ld = which_binary('ld')
        if host_ld:
            ld_version = check_output([host_ld, '-version'])
            if ld_version.count('gold') == 1:
                extra_host_ldflags += '-fuse-ld=bfd '

        set_config('EXTRA_TARGET_LDFLAGS', extra_target_ldflags)
    set_config('EXTRA_HOST_LDFLAGS', extra_host_ldflags)


def pkg_config():
    '''
    If package configuration is enabled, then for each library in PKG_CONFIG_PACKAGES, the
    pkg-config utility will be invoked to populate configuration variables.
    The cflags, linker paths and libraries will be assigned to XXX_CFLAGS, XXX_LDFLAGS
    and XXX_LIBS respectively, where XXX is the uppercase package name with any non
    alphanumeric letters replaced by '_'.
    Where no package information exists the default configuration value will be used.
    '''
    if get_config('PKG_CONFIG')['value'] == 'y':
        pkg_config_path = get_config('PKG_CONFIG_PATH')['value']
        if pkg_config_path != '':
            os.putenv('PKG_CONFIG_PATH', pkg_config_path)

        pkg_config_sys_root = get_config('PKG_CONFIG_SYSROOT_DIR')['value']
        if pkg_config_sys_root != '':
            os.putenv('PKG_CONFIG_SYSROOT_DIR', pkg_config_sys_root)

        pkg_config_packages = get_config('PKG_CONFIG_PACKAGES')['value']

        pkg_config_packages_list = pkg_config_packages.split(',')

        for pkg in pkg_config_packages_list:
            pkg = pkg.strip()
            if pkg == '': continue
            # convert library name to upper case alpha numeric
            pkg_uc_alnum = re.sub('[^a-zA-Z0-9_]', '_', pkg.upper())

            pkg_config_cflags = "%s%s" % (pkg_uc_alnum, '_CFLAGS' )
            pkg_config_ldflags = "%s%s" % (pkg_uc_alnum, '_LDFLAGS' )
            pkg_config_libs = "%s%s" % (pkg_uc_alnum, '_LDLIBS' )

            cflags = check_output(['pkg-config', pkg, '--cflags'])
            if cflags != '':
                set_config(pkg_config_cflags, cflags)

            ldflags = check_output(['pkg-config', pkg, '--libs-only-L'])
            if ldflags != '':
                set_config(pkg_config_ldflags, ldflags)

            libs = check_output(['pkg-config', pkg, '--libs-only-l'])
            if libs != '':
                set_config(pkg_config_libs, libs)



def plugin_exec():
    if get_config('ALLOW_HOST_EXPLORE')['value'] == 'y':
        compiler_config()
        pkg_config()
