# -----------------------------------------------------------------------------
# This confidential and proprietary software may be used only as
# authorised by a licensing agreement from ARM Limited
# (C) COPYRIGHT 2010-2011, 2013-2018 ARM Limited
# ALL RIGHTS RESERVED
# The entire notice above must be reproduced on all authorised
# copies and copies may only be made to the extent permitted
# by a licensing agreement from ARM Limited.
# -----------------------------------------------------------------------------
__docformat__ = "restructuredtext en" 

import os, sys, subprocess

from os_abstr import get_appropriate_os, get_build_os
from toolchain_abstr import get_appropriate_toolchain, get_gcc_version
from platform_abstr  import get_appropriate_platform, get_build_platform

# =============================================================================

class BaseEnvironmentConfig( object ):

	"""
	This class implements a base class from which other environments are derived.
	"""

	# -------------------------------------------------------------------------
	def __init__( self ):

		self.rBuildDir = ""
		self.rBuildDirHash = ""
		self.rVariant = ""

	# -------------------------------------------------------------------------
	def construct_variant_name( self, sEnv, lSupportedOptions ):
		'''
		Sets ``self.rBuildDir`` to the name (str) of the current variant.

		Consider only the host platform, and options marked as "build" in the
		list of supported options. When a build tOptions is found, check the
		environment to see if it is present.  If it is present, add it & its
		value to the list of atoms. Once all supported options have been
		checked, assemble the variant name from the list of atoms.

		:Parameters:
			sEnv : `SConcsEnvironment`
				SCons state
			lSupportedOptions : list
				Candidate options to consider as part of the hash. Will
				subject to filtering, so only build options are used in the
				calculation.Sequence of::

					(str,
						(str, str, int, callable, callable)
					)
		'''
		from zlib import crc32
		import platform

		rPlatformID = " ".join( platform.dist() )

		lAtoms = []
		lAtoms.append("$PLATFORM_DIST="+rPlatformID)
		for rOpt, tOptions in sorted( lSupportedOptions ):
			if rOpt != "build":
				# This option doesn't affect build output
				continue

			rFlag = tOptions[0]
			if sEnv.has_key(rFlag):
				lAtoms.append( '%s=%s' % (rFlag, sEnv[rFlag]) )

		self.rVariant = " ".join( lAtoms )
		self.rBuildDirHash = "%x" % abs( crc32(self.rVariant) )

		self.rBuildDir = os.path.join( 'build', self.rBuildDirHash )

		if not os.path.exists( 'build' ):
			os.makedirs( 'build' )

		rVariantFile = str("%s.variant.txt" % self.rBuildDir,)

		with file(rVariantFile, "w") as sFH:
			sFH.write( self.rVariant )
			sFH.write( "\n" )

	# -------------------------------------------------------------------------
	def update_environment( self, sEnv, rSrcDir ):

		rStaticLibPath = os.path.join( os.getcwd(), os.path.join( self.rBuildDir, 'slib' ) )
		sEnv.Append( STATIC_LIB_PATH=rStaticLibPath )

		rSharedLibPath = os.path.join( os.getcwd(), os.path.join( self.rBuildDir, 'dlib' ) )
		sEnv.Append( SHARED_LIB_PATH=rSharedLibPath )

		rSharedStubLibPath = os.path.join( os.getcwd(), os.path.join( self.rBuildDir, 'dlib_stub' ) )
		sEnv.Append( SHARED_STUB_LIB_PATH=rSharedStubLibPath )

		rUnitTestPath = os.path.join( os.getcwd(), os.path.join( self.rBuildDir, 'unit' ) )
		sEnv.Append( UNITTEST_PROGS_PATH=rUnitTestPath )

		rPerfTestPath = os.path.join( os.getcwd(), os.path.join( self.rBuildDir, 'perf' ) )
		sEnv.Append( PERFTEST_PROGS_PATH=rPerfTestPath )

		rBuildDirPath = os.path.join( os.getcwd(), self.rBuildDir )
		sEnv.Append( BUILD_DIR_PATH=rBuildDirPath )

		if sEnv['os'] == 'android':
			sEnv['ANDROID_PREBUILT_DIR'] = os.path.join(sEnv.GetLaunchDir(), 'android_prebuilt')
			if sEnv['standalone'] == '1':
				sEnv['PREBUILT_DIR'] = sEnv['BUILD_DIR_PATH']
			else:
				sEnv['PREBUILT_DIR'] = sEnv['ANDROID_PREBUILT_DIR']
		else:
			sEnv['PREBUILT_DIR'] = sEnv['BUILD_DIR_PATH']

		# The following block constructs the MALI_VERSION_INFO define
		rMaliReleaseName = sEnv['ENV']['MALI_RELEASE_NAME']

		# Set MALI_BUILD_VARIANT to our build variant (Massage the string to remove
		# linefeeds, replace all spaces with underscore and join list of words to a
		# single string) Doing this prevents build issues with Android.
		if sEnv['release'] == '1':
			sEnv['MALI_BUILD_VARIANT'] = "Release"
		else:
			sEnv['MALI_BUILD_VARIANT'] = str("".join(str(self.rVariant).replace('\n', ' ').\
				replace(' ', '_').split()))

		# check MALI_RELEASE_NAME is default value, if yes, assume we are in a git repo
		if "not_released" == rMaliReleaseName:
			# get the sha1 of HEAD
			try:
				lGitCmd = ['git','rev-parse','--short','--verify','HEAD']
				sGitProc = subprocess.Popen(lGitCmd,
					stdout = subprocess.PIPE, 
					stderr = subprocess.PIPE)
				rStdout,rStderr = sGitProc.communicate()

				if ( (sGitProc.returncode != 0) or rStderr or ( not rStdout) ):
					raise RuntimeError()
				rGitShaShort = rStdout.strip()

			except Exception:
				sys.stdout.write("bldsys warning : Unable to get Git revision for MALI_VERSION_INFO\n")
				rGitShaShort = "git_rev_unknown"

			# Prepare MALI_VERSION_INFO to contain "not_released.<git-sha-short>" - the build hash follows after scanning files.
			sEnv['MALI_VERSION_INFO'] = rMaliReleaseName + "." + rGitShaShort
		else:
			# Prepare MALI_VERSION_INFO to contain "v1.<release_name>" - the build hash follows after scanning files.
			sEnv['MALI_VERSION_INFO'] = "v1." + rMaliReleaseName


	# -------------------------------------------------------------------------
	def apply_os_settings(self, sEnv):
		sOS = get_appropriate_os( sEnv )
		sOS.set_windowing_system( sEnv['winsys'] )
		if isinstance(self, TargetEnvironmentConfig):
			sOS.set_as_target()
		sOS.update_environment( sEnv )

	# -------------------------------------------------------------------------
	def apply_platform_settings(self, sEnv):
		sPlatform = get_appropriate_platform( sEnv )
		sPlatform.update_environment( sEnv )

	# -------------------------------------------------------------------------
	def apply_toolchain_settings(self, sEnv):

		sToolchain = get_appropriate_toolchain( sEnv )
		if isinstance(self, HostEnvironmentConfig) == True:
			sToolchain.set_toolchain_prefix( sEnv['toolchain_prefix_host'] )
		if isinstance(self, TargetEnvironmentConfig) == True:
			sToolchain.set_toolchain_prefix( sEnv['toolchain_prefix_target'] )
		sToolchain.set_optimization_flags( sEnv )
		# Debug flags need to be set after optimization flags so that stack frames flags will
		# override the -O flag.
		sToolchain.set_debug_symbol_flags( sEnv['symbols'], sEnv['sanitizer'], sEnv )
		if sEnv.has_key( 'exclude' ) != False:
			sToolchain.set_exclude_flags( sEnv['exclude'] )

		# Check the GCC version. We don't use the cross-compiler to build target user-side
		# code on Android builds as we get the Android tree to do the build using its own tools,
		# consequently we only check the version of GCC for linux and host side Android.
		if (isinstance(self, HostEnvironmentConfig) == True or "linux" == sEnv['os']) and sEnv['toolchain'] == 'gnu':
			(iMajor, iMinor, iPatch) = get_gcc_version( sToolchain.rCC )
			if ((iMajor < 4) or ((iMajor == 4) and (iMinor < 9))):
				sys.stderr.write( "bldsys error: Support for building with versions of gcc older than 4.9.x for %s has been removed (found %d.%d.%d).\n" % (sToolchain.rTargetArch, iMajor, iMinor, iPatch) )
				sys.exit(1)

		sToolchain.configure_linker( sEnv )
		sToolchain.update_environment( sEnv )

# =============================================================================
class HostEnvironmentConfig( BaseEnvironmentConfig):

	"""
	This class implements an instance of the environment abstraction for the host.
	"""

	# -------------------------------------------------------------------------
	def __init__( self ):

		BaseEnvironmentConfig.__init__(self)

	# -------------------------------------------------------------------------
	def add_build_dir_suffix( self ):
	
		self.rBuildDir = self.rBuildDir +'-host' 		

	# -------------------------------------------------------------------------
	def apply_os_settings(self, sEnv):
		rBuildOS = get_build_os()
		sEnv['os'] = rBuildOS
		BaseEnvironmentConfig.apply_os_settings( self, sEnv )

	# -------------------------------------------------------------------------
	def apply_platform_settings(self, sEnv):
		rBuildArch = get_build_platform()
		sEnv['arch'] = rBuildArch
		BaseEnvironmentConfig.apply_platform_settings( self, sEnv )

# =============================================================================
class TargetEnvironmentConfig( BaseEnvironmentConfig ):

	"""
	This class implements an instance of the environment abstraction for the target.
	"""

	# -------------------------------------------------------------------------
	def __init__( self ):

		BaseEnvironmentConfig.__init__(self)

	def apply_toolchain_settings( self, sEnv ):

		BaseEnvironmentConfig.apply_toolchain_settings ( self, sEnv )

		try:
			# honor CFLAGS from system environment
			cflags = os.environ['CFLAGS']
			cxxflags = os.environ['CXXFLAGS']
			if ('-march' in cflags or '-mcpu' in cflags or '-mtune' in cflags
				or '-march' in cxxflags or '-mcpu' in cxxflags or '-mtune' in cxxflags):
					for flag in sEnv['CCFLAGS']:
						if '-march' in flag or '-mcpu' in flag or '-mtune' in flag:
							sEnv['CCFLAGS'].remove(flag);
			sEnv['CFLAGS'] += cflags.split()
			sEnv['ASFLAGS'] += cflags.split()
			sEnv['CXXFLAGS'] += cxxflags.split()
			sEnv['CCFLAGS'] += os.environ['CPPFLAGS'].split()
		except:
			pass

	# -------------------------------------------------------------------------
	def add_build_dir_suffix( self ):
		
		self.rBuildDir = self.rBuildDir + '-target' 		
