# This confidential and proprietary software may be used only as
# authorised by a licensing agreement from ARM Limited
# (C) COPYRIGHT 2010-2012, 2014-2017 ARM Limited
# ALL RIGHTS RESERVED
# The entire notice above must be reproduced on all authorised
# copies and copies may only be made to the extent permitted
# by a licensing agreement from ARM Limited.

import os, sys, functools
from collections import namedtuple
import pipes

# need to import here as they're not added to the environment by default as
# we're importing not execing this file.
from SCons.Script.Main import AddOption, GetOption
from SCons.Script import Glob
from SCons.Node import Node


class ExternalLibraryFixup(namedtuple('ExternalLibraryFixup', [
		'min_version',   # Minium version of the library that the fixup is
		                 # required for. If None then the version is unbounded.
		'max_version',   # Minium version of the library that the fixup is
		                 # required for. If None then the version is unbounded.
		'fixup_function' # fixup_function(env) updates the passed environment
		                 # with the fixups required for the library versions
		                 # specified
	])):
	'''
	A Class to encapsulate any environment fixes required when including
	a dependency on a third party library. For example this could relax
	the build for versions of a library where the libraries headers would
	normally cause a build failure.
	'''
	pass

class ExternalLibraryInfo(namedtuple('ExternalLibraryInfo', [
		'lib_name',   # Library name to link against (if pkg-config unavailable)
		'local_src',  # function returns True if the library is available in
		              # tree, and therefore pkg-config should *not* be used
		'min_version',# Minimum version required of the library, or None
		              # if the version should not be checked
		'checks'      # Set of ExternalLibraryFixup instances that define any
		              # fixups required for this library
	])):
	'''
	A Class to describe an external library that may be required by parts
	of the DDK build. It includes all of the information required to add
	a dependency on the library to the current environment.
	'''
	pass

# A dict of ExternalLibraryInfo to describe how to include all of the external
# libraries that may need to be included for the different winsys
# implementations to build. The keys are the pkg-config library names.
#
ext_lib_info = {
	'libdrm' : ExternalLibraryInfo(
		lib_name = 'drm',
		local_src = lambda : any([Glob('#' + winsys + '/thirdparty/include/libdrm') for winsys in ['wayland','gbm','vulkan_wsi']]),
		min_version = '2.4.34',
		checks = set([ExternalLibraryFixup(
			min_version = None,
			max_version = '2.4.46',
			fixup_function = lambda env: remove_cflags(env, ['-pedantic']))])),
	'gbm' : ExternalLibraryInfo(
		lib_name = None,           # Header only
		local_src = lambda : True, # Currently can't use pkg-config, so assume
		                           #it is always present locally
		min_version = None,
		checks = set([ExternalLibraryFixup(
			min_version = None,
			max_version = None,
			fixup_function = lambda env: remove_cflags(env, ['-pedantic']))])),
	'wayland-client' : ExternalLibraryInfo(
		lib_name = 'wayland-client',
		local_src = lambda : Glob('#wayland/thirdparty'),
		min_version = '1.5.0',
		checks = set([ExternalLibraryFixup(
			min_version = None,
			max_version = None,
			fixup_function = lambda env:
				add_remove_cflags(env,
				                   ['-Wno-deprecated-declarations'],
				                   ['-pedantic']))])),
	'wayland-server' : ExternalLibraryInfo(
		lib_name = 'wayland-server',
		local_src = lambda : Glob('#wayland/thirdparty'),
		min_version = '1.5.0',
		checks = set([ExternalLibraryFixup(
			min_version = None,
			max_version = None,
			fixup_function = lambda env:
				add_remove_cflags(env,
				                   ['-Wno-deprecated-declarations'],
				                   ['-pedantic']))]))
	}

AddOption('--profiles-paths',
	dest='profiles_paths',
	action='store',
	default='',
	help='Extra paths to search for profiles in. (e.g. --profile-paths=~/myprofiles:/work/profiles)')

# split, strip and remove empty paths
user_paths = GetOption('profiles_paths').split(':')
user_paths = [os.path.expanduser(path).strip() for path in user_paths]
user_paths = [path for path in user_paths if path != '']

# user paths take priority (are searched first)
# sys.path[0] give us the directory containing this script
profiles_paths = user_paths + [ os.path.join(sys.path[0], 'profiles') ]
if os.path.exists(os.path.join(sys.path[0], 'profiles', 'deprecated')):
	profiles_paths += [os.path.join(sys.path[0], 'profiles', 'deprecated')]

def get_available_profiles():
	'''
	Helper to determine the available profiles.
	Used by the top level sconstruct when generating the online help display.
	'''

	available_profiles = []

	for profiles_path in profiles_paths:
		if not os.path.isdir(profiles_path):
			sys.stderr.write('*** Profile directory %s doesn\'t exist.\n' % profiles_path)
		else:
			for p in os.listdir(profiles_path):
				name, ext = os.path.splitext(p)
				if '.py' == ext:
					if name in available_profiles:
						sys.stderr.write('*** Conflicting profile %s\n' % p)
					else:
						available_profiles.append( name )

	return available_profiles


def load_profiles(args=[]):
	'''
	Helper to process all 'profile=value' arguments into a list
	Used by the top level sconstruct.
	'''
	profiles = []
	for k, v in args:
		if 'profile' == k:
			for profiles_path in profiles_paths:
				fname = os.path.join(profiles_path, '%s.py' % v)
				if os.path.isfile(fname):
					try:
						os.stat( fname )
						profiles.append( fname )
					except StandardError as ex:
						sys.stderr.write('*** Failed to load profile "%s":\n\n%s\n' % (v, ex))
						exit(1)
					if os.path.basename(os.path.dirname(fname)) == "deprecated":
						sys.stderr.write("WARNING - Profile '%s' is deprecated and will be removed!\n" % v)
	return profiles


def lib_target(env, name, source, *args, **kw):
	'''
	Helper to expose a library target as alias that can be specified on
	the command line.
	Added to the environment as LibTarget.

	This helper takes care of correctly creating and installing version symlinks
	if built with soversion set. For example, if env['soversion']=(1,4,0),
	env['libs_install'] = 'libs' and the following is called:

		lib = env.SharedLibrary(target = '$SHARED_LIB_PATH/libEGL')
		env.LibTarget('egl', lib)

	then the following will be installed into libs/:

		libEGL.so.1.4.0 (the source library))
		libEGL.so -> libEGL.so.1.4.0
		libEGL.so.1 -> libEGL.so.1.4.0
		libEGL.so.1.4 -> libEGL.so.1.4.0

	(where the arrows represent symlinks)
	'''

	#source may be an iterable: lets make it always a list to make things a bit simpler
	try:
		iter(source)
	except TypeError, te:
		source = [source]

	if env.has_key('libs_install'):
		installed_files = []
		if env.has_key('soversion') and env['soversion'] is not None:

			for src in source:
				# Only shared libraries are versioned - I don't know of a clean and
				# stable way to determine if a node is a shared library other than
				# checking for the .so suffix?
				if str(src).endswith('.so'):
					name_no_suffix = os.path.join(env['libs_install'],os.path.basename(str(src)))
					name_major_suffix = '{}.{}'.format(name_no_suffix, env['soversion'][0])
					name_minor_suffix = '{}.{}'.format(name_major_suffix, env['soversion'][1])
					name_patch_suffix = '{}.{}'.format(name_minor_suffix, env['soversion'][2])

					i = env.InstallAs(name_patch_suffix, src)
					installed_files += i
					installed_files += env.SymLink(name_no_suffix, i)
					installed_files += env.SymLink(name_major_suffix, i)
					installed_files += env.SymLink(name_minor_suffix, i)
				else:
					installed_files += env.Install(env['libs_install'], src)
		else:
			installed_files += env.Install(env['libs_install'], source)

		env.Alias(name, installed_files)
	else:
		env.Alias(name, source)

	# Regardless of whether we are installing or not we create an SONAME'd symlink
	# in SHARED_LIB_PATH, for use by the tests built with the DDK
	# However, SCons doesn't check the SONAME when adding dependencies based on
	# a program or libraries LIBS, so won't ensure the links for any internal
	# shared objects are created before they are linked.
	# We therefore have to reverse the dependency, to force the link to be created
	# before the library it links to.
	if env.has_key('soversion') and env['soversion'] is not None:
		for src in source:
			if str(src).endswith('.so'):
				symlink = env.SymLinkNoDep("${SHARED_LIB_PATH}/${SOURCE.file}.${soversion[0]}", src)
				env.Depends(src, symlink)


def resource_target(env, folder_name, source, alias = None, *args, **kw):
	'''
	Helper to install resources to correct folder with possible alias.
	This feature defaults to the behavior of progs_install when
	resource_install is not specified.
	'''
	if 'resource_install' in env:
		i = env.Install( os.path.join(env['resource_install'], folder_name), source)
	elif 'progs_install' in env:
		i = env.Install( os.path.join(env['progs_install'], folder_name), source)
	else:
		i = env.Install( os.path.join('$UNITTEST_PROGS_PATH', folder_name), source)

	if (alias != None):
		env.Alias(alias, i)
	else:
		env.Alias(folder_name, i)

	return i

def kernel_obj_target(env, folder_name, source, alias = None, *args, **kw):
	'''
	Helper to install kernel objects to correct folder with possible alias.
	At the moment, we are installing kernel modules into the folder specified
	by resource_install build option. This feature defaults to the behavior of
	progs_install when resource_install is not specified.
	'''
	if 'resource_install' in env:
		i = env.Install(env['resource_install'], source)
	elif 'progs_install' in env:
		i = env.Install(env['progs_install'], source)
	else:
		i = env.Install('$UNITTEST_PROGS_PATH', source)

	if (alias != None):
		env.Alias(alias, i)
	else:
		env.Alias(folder_name, i)

	return i

def prog_target(env, name, source, *args, **kw):
	'''
	Helper to expose a program target as alias that can be specified on
	the command line.
	Added to the environment as ProgTarget.
	'''

	#Since, progs are compiled later in android, we bail out early.
	if env['winsys'][0] == 'android':
		env.Alias(name, source)
		return
	if 'progs_install' in env:
		i = env.Install(env['progs_install'], source)
		env.Alias(name, i)
	else:
		env.Alias(name, source)

def compiler_tool(env, name, source, *args, **kw):
	'''
	Helper to expose a program target as alias that can be specified on
	the command line.
	Added to the environment as CompilerTool.
	'''

	if 'progs_install' in env:
		i = env.Install(env['progs_install'], source)
	else:
		i = env.Install(os.path.join(env.GetLaunchDir(), 'compiler_tools'), source)
	env.Alias(name, i)

def info_print(env, s):
	'''
	Print informational messages, respecting the verbose (v) command line
	switch. Used within the bldsys module only.
	'''
	if env['v'] == '1':
		print s

import collections
import SCons.Util

class _OrderedListVariable(collections.UserList):
	'''
	This class and associated functions (including _converter and OrderedListVariable)
	is derived from the SCons ListVariable code located here:
	http://www.scons.org/doc/2.0.1.beta.20100627/HTML/scons-api/SCons.Variables.ListVariable%27-pysrc.html

	The original SCons ListVariable behaves similar to a Set rather than a list and
	does not maintain the order of the variables specified on the command line. This
	list variable type maintains the order specified during creation.
	'''
	def __init__( self, initlist=[] ):
		collections.UserList.__init__(self, [_f for _f in initlist if _f])

	def __cmp__(self, other):
		raise NotImplementedError
	def __eq__(self, other):
		raise NotImplementedError
	def __ge__(self, other):
		raise NotImplementedError
	def __gt__(self, other):
		raise NotImplementedError
	def __le__(self, other):
		raise NotImplementedError
	def __lt__(self, other):
		raise NotImplementedError
	def __str__( self ):
		return '_'.join( self )
	def prepare_to_store(self):
		return self.__str__()

def _converter( val, allowedElems, mapdict ):
	'''
	This converter maps the values given during creation
	to new values and validates them. If no map was specified
	then the original values are validated and used.
	'''
	val = [_f for _f in val.split(',') if _f]
	val = [mapdict.get(v, v) for v in val]
	notAllowed = [v for v in val if not v in allowedElems]
	if notAllowed:
		raise ValueError("Invalid value(s) for option: %s" %
				 ','.join(notAllowed))
	return _OrderedListVariable(val)

def OrderedListVariable( key, help, default, names, map={} ):
	'''
	The input parameters describe a list of variable names and are
	returned with the correct converter appended. The result is
	usable for input to opts.Add().

	Unlike ListVariables, OrderedListVariables do not support the
	'all' or 'none' value.
	'''
	names_str = 'allowed names: %s' % ' '.join( names )
	if SCons.Util.is_List(default):
		default = ','.join(default)
	help = '\n    '.join(
		(help, '(comma-separated list of names)', names_str))
	return (key, help, default,
		None, lambda val: _converter(val, names, map))


def expand_list(env , varList):
	'''
	Expands an environment variable or list into a list of values.
	Similar to env.subst but does not need splitting.
	Similar to (undocumented) env.subst_list but does not
	return nested list.

	WARNING: If the sconscript in which this is used is not read
	after the definition of environment variables then the references
	are lost, so needed values may be missed.
	'''
	# Substitutes each list element with expanded form which are split into
	# sublists. sum(_,[]) is used to concatenate the result.
	return sum([env.subst(listItem).split() for listItem in varList],[])

def remove_cflags(env, flag_list):
	'''
	Removes all flags in flag_list from env['CFLAGS'] (if present)
	'''
	# NB: env[..].remove is not working on scons < 2.0.1, so copy and replace
	cflags = env['CFLAGS'][:]
	for flag in flag_list:
		if flag in cflags:
			cflags.remove(flag)
	env.Replace(CFLAGS = cflags)

def add_remove_cflags(env, list_to_add, list_to_remove):
	'''
	Adds all flags in list_to_add to env['CFLAGS'], and removes any flags in list_to_remove (if present)
	'''
	# NB: env[..].remove is not working on scons < 2.0.1, so copy and replace
	cflags = env['CFLAGS'][:]
	for flag in list_to_add:
		cflags.append(flag)
	env.Replace(CFLAGS = cflags)
	remove_cflags(env, list_to_remove)

def check_pkg_config(env, ctx, version=0):
	ctx.Message( 'Checking for pkg-config...' )
	cmd = '{0} --atleast-pkgconfig-version={1}'.format(env['ENV']['PKG_CONFIG'], version)
	ret = ctx.TryAction(cmd)[0]
	ctx.Result(ret)
	return ret

lPkgConfigChecked = set()
def check_pkg(env, ctx, pkg, min_version=None, max_version=None):
	lPkgConfigChecked.add(pkg)
	ret=True

	if min_version is None and max_version is None:
		ctx.Message( 'Checking for {}...'.format(pkg) )
		cmd = '{0} --exists {1}'.format(env['ENV']['PKG_CONFIG'], pkg)
		ctx.Result(ctx.TryAction(cmd)[0])
		ret = ctx.TryAction(cmd)[0]
	else:
		if min_version is not None:
			ctx.Message( 'Checking for {0} >= {1}...'.format(pkg, min_version) )
			cmd = '{0} --atleast-version={1} {2}'.format(env['ENV']['PKG_CONFIG'], min_version, pkg)
			ctx.Result(ctx.TryAction(cmd)[0])
			ret = ctx.TryAction(cmd)[0]

		if max_version is not None:
			ctx.Message( 'Checking for {0} <= {1}...'.format(pkg, max_version) )
			cmd = '{0} --max-version={1} {2}'.format(env['ENV']['PKG_CONFIG'], max_version, pkg)
			ctx.Result(ctx.TryAction(cmd)[0])
			ret = ret and ctx.TryAction(cmd)[0]

	return ret

def get_configure(env):
	'''
	Gets a Scons Configure instance, using a common directory and log file
	within the build directory
	'''
	confDir = os.path.join(env['BUILD_DIR_PATH'], 'sconf_temp')
	dCustomTests = { 'CheckPkgConfig' : functools.partial(check_pkg_config, env),
	                 'CheckPkg'       : functools.partial(check_pkg, env) }
	return env.Configure(conf_dir=confDir, log_file=os.path.join(confDir, 'config.log'),
	                     clean=False, help=False, custom_tests = dCustomTests)

def pkg_config_add(sEnv, lPkgs, yAddLibs = True):
	# Don't add packages when cleaning
	if sEnv.GetOption('clean'):
		return

	rLibArg = '--libs' if yAddLibs else ''

	for rPkg in lPkgs:
		if sEnv['release'] == '0':
			# Check that we have verified the pkg is present before trying
			# to parse
			assert rPkg in lPkgConfigChecked

		sEnv.ParseConfig('{0} --cflags {1} {2}'.format(sEnv['ENV']['PKG_CONFIG'], rLibArg, rPkg))

def filtered_glob(self, pattern, omit=[], ondisk=True, source=False, strings=False):
	"""
	Filter a Glob by omitting the list of file names in `omit`.  Intended to be
	attached to a Scons ``env`` object, so ``self`` would quack like an ``env``.

	Most arguments except for `omit` will be passed through to the ``scons`` ``Glob``
	function, so for more documentation, consult the Scons documentation.  For example,
	in a Unix environment with ``scons`` installed, type ``man scons``.

	:Parameters:
		pattern : ``str``
			Pattern to match filenames against.
		omit : ``list`` of ``str``
			List of files to exclude, after filtering according to `pattern`.  Every element of
			this list should correspond to the result of ``os.path.basename(f.path)`` where
			``f`` is a ``SCons.Node``.
		ondisk : ``bool``
			Set to ``False`` (or any other non-true value) to disable the search for
			matches on disk, thereby only returning matches among already-configured
			``File`` or ``Dir`` ``Nodes``.
		source : ``bool``
			May be set to ``True`` (or any equivalent value) to specify that, when the
			local directory is a ``VariantDir()``, the returned ``Node`` instances
			should be from the corresponding source directory, not the local directory.
		strings : ``bool``
			May be set to ``True`` (or any equivalent value) to have the ``Glob()`` function
			return strings, not ``Node`` instances, that represent the matched files or
			directories.

	:Returns:
		``Node`` instances (or strings) that match the specified `pattern`, relative to the directory
		of the current ``Sconscript`` file.
	"""
	if strings:
		return filter(
			lambda f: os.path.basename(f) not in omit,
			self.Glob(pattern, ondisk, source, strings))
	else:
		return filter(
			lambda f: os.path.basename(f.path) not in omit,
			self.Glob(pattern, ondisk, source, strings))

# Dictionary that gets populated by calls to add_ext_libs_to_env to record
# which libraries have had their versions checked, and what fixups to
# the build are required
ext_lib_fixups = dict()

def add_ext_libs_to_env(sEnv, lPkgs, yAddLibs = True):
	'''
	Updates the environment to include the specified pkg-config pkgs,
	ensuring an error-free build by relaxing the build options as required.
	This function triggers lazy version checks, to ensure all libraries
	are either present within the DDK tree, or are found by pkg-config.
	'''

	# True if the min_versions of the packages should be checked - we
	# do this even if the package is included in-tree, so we can prevent
	# regressions while using filesystems that cannot be used to
	# cross-compile
	yCheckPkgVersions = ( sEnv['ENV']['PKG_CONFIG'] != 'pkg-config' or
	                      'PKG_CONFIG_PATH' in sEnv['ENV'] )

	sConf = sEnv.GetConfigure()

	for rPkg in lPkgs:
		if rPkg not in ext_lib_info:
			print 'WARNING: Including a package for which no information is recorded.'
			print 'No version checks or build fixups wil be recorded, and libraries'
			print 'maybe incorrectly linked'
			print '  Package requested : ' + rPkg
			if sEnv['release'] == '0':
				Exit(1) # bail, to ensure the information is added
			sLibInfo = ExternalLibraryInfo(rPkg, lambda : True, None, set())
		else:
			sLibInfo = ext_lib_info[rPkg]

		if rPkg in ext_lib_fixups:
			# Library has been seen before, so just run any fixups that were
			# recorded for it
			if ext_lib_fixups[rPkg]:
				ext_lib_fixups[rPkg](sEnv)
		else:
				if yCheckPkgVersions and sLibInfo.min_version:
					if not sConf.CheckPkg(rPkg, sLibInfo.min_version):
						sEnv.Replace( MissingPkgs = True )

				sFixupFuncs = set()
				for sCheck in sLibInfo.checks:
					if ( sLibInfo.local_src() or
					     sConf.CheckPkg(rPkg, sCheck.min_version, sCheck.max_version) ):
						sFixupFuncs.add(sCheck.fixup_function)

				def fFixup(sEnv):
					for fFunc in sFixupFuncs:
						fFunc(sEnv)

				ext_lib_fixups[rPkg] = fFixup
				fFixup(sEnv)

		if sLibInfo.local_src():
			if yAddLibs and sLibInfo.lib_name:
				sEnv.Append(LIBS = [sLibInfo.lib_name])
		else:
			sEnv.PkgConfigAdd([rPkg], yAddLibs)

	sConf.Finish()

def get_ext_lib_deps(env):
	'''
	Looks through the LIBS environment to reverse-engineer which pkg-config
	packages were used to include the libraries. The result is a string
	that can be used for a .pc "Requires.private:" keyword
	'''
	libs = env.Flatten(env['LIBS'])
	dep_str = []
	for pkg, info in ext_lib_info.items():
		if not info.local_src() and info.lib_name in libs:
			dep_str.append("pkg")
			if info.min_version:
				dep_str.append(">=")
				dep_str.append(info.min_version)
	return " ".join(dep_str)

def find_c_dependencies(target):
	'''
	Walks the dependency graph of target to find all .c and .h files it depends on.
	'''
	def _find_sources(tgt, srcs, seen):
		for item in tgt:
			if SCons.Util.is_List(item):
				_find_sources(item, srcs, seen)
			elif isinstance(item, SCons.Node.FS.Base) and item.abspath not in seen:
				seen[item.abspath] = True

				if (item.path.endswith('.c') or item.path.endswith('.h') or
						item.path.endswith('.cpp') or item.path.endswith('.hpp')):
					i = item
					while not i.exists() and i != i.srcnode():
						i = i.srcnode()
					if i.exists():
						srcs.append(i)
				_find_sources(item.children(scan=1), srcs, seen)

	sources = []
	_find_sources(target, sources, {})
	sources.sort(key=Node.get_abspath)
	return sources

def symlink_action(target=None, source=None, env=None):
	if isinstance(target, list):
		target = target[0]

	src_str = str(source[0])
	dst_str = str(target)

	dst_dir = os.path.dirname(dst_str)
	src_str = os.path.relpath(src_str, dst_dir if dst_dir else os.curdir)

	os.symlink(src_str, dst_str)

def pkgconfig_helper(env, source, subst_dict = dict(), target=None):
	'''
	Helper to add a builder of the pkg-config .pc file named,
	using the source as the template file.
	Additional keyword arguments will be used as the substitions to be made.
	'''

	if 'pkgconfig_install' not in env:
		return

	subst_dict.setdefault('@prefix@', '$pkgconfig_prefix')
	subst_dict.setdefault('@version@', "${soversion[0]}.${soversion[1]}.${soversion[2]}")

	if target:
		pc_file = env.Substfile(target, source, SUBST_DICT = subst_dict)
	else:
		pc_file = env.Substfile(source, SUBST_DICT = subst_dict)

	env.Install(env['pkgconfig_install'], pc_file)

def quote_command(cmd):
	return " ".join([pipes.quote(i) for i in cmd])
